import os
import lucene
import time

from java.io import File
from java.nio.file import Paths
from org.apache.pylucene.analysis import PythonAnalyzer
from org.apache.lucene.analysis import LowerCaseFilter, StopFilter
from org.apache.lucene.analysis.standard import StandardAnalyzer, StandardTokenizer
from org.apache.lucene.analysis.pt import PortugueseAnalyzer, PortugueseLightStemFilter
from org.apache.lucene.document import Document, Field, TextField, StringField
from org.apache.lucene.search import IndexSearcher, FuzzyQuery
from org.apache.lucene.index import DirectoryReader, Term
from org.apache.lucene.queryparser.classic import QueryParser
from org.apache.lucene.index import IndexWriter, IndexWriterConfig
from org.apache.lucene.store import FSDirectory, SimpleFSDirectory
from org.apache.lucene.util import Version

from ASAPPpy import ROOT_PATH

class customPortugueseAnalyser(PythonAnalyzer):

	def createComponents(self, fieldName):
		source = StandardTokenizer()
		stream = LowerCaseFilter(source)
		stream = StopFilter(stream, PortugueseAnalyzer.getDefaultStopSet())
		stream = PortugueseLightStemFilter(stream)

		return self.TokenStreamComponents(source, stream)

	def initReader(self, fieldName, reader):
		return reader

def l_indexer(directory, load_path):
	lucene.initVM()

	# index_dir = SimpleFSDirectory(File(directory))
	index_dir = FSDirectory.open(Paths.get(directory))
	writer_config = IndexWriterConfig(PortugueseAnalyzer())
	# writer_config = IndexWriterConfig(customPortugueseAnalyser())
	writer = IndexWriter(index_dir, writer_config)

	with open(load_path) as subtles_file:
		subtles_corpus = subtles_file.read().splitlines()

	for i in range(0, len(subtles_corpus), 2):
		doc = Document()
		doc.add(Field("question", subtles_corpus[i], StringField.TYPE_STORED))
		doc.add(Field("answer", subtles_corpus[i+1], StringField.TYPE_STORED))

		writer.addDocument(doc)

	writer.close()
	print("Index successfully created!")

def l_searcher(query_string, directory, number_documents):
	lucene.initVM()

	# analyzer = StandardAnalyzer()
	reader = DirectoryReader.open(FSDirectory.open(Paths.get(directory)))
	searcher = IndexSearcher(reader)

	# Top 'n' documents as result
	topN = number_documents

	try:
		# query = QueryParser("question", analyzer).parse(query_string)
		query = FuzzyQuery(Term("question", query_string), 2)
		print("The query was: {}".format(query))

		hits = searcher.search(query, topN)

		print("The hits were: ")

		options = []
		options_answers = []

		# print(hits.totalHits)

		for hit in hits.scoreDocs:
			print(hit.doc)
			# print(hit.score, hit.doc, hit.toString())
			doc = searcher.doc(hit.doc)
			options_answers.append(doc.get("answer"))
			options.append(doc.get("question"))
			# print(doc.get("answer"))

		return options, options_answers
	except IndexError:
		return None

if __name__ == '__main__':
	# load_path = os.path.join(ROOT_PATH, 'datasets', 'SubtleCorpusPTEN', 'por', 'corpus0sDialogues_clean_removed_entities.txt')
	# directory = os.path.join('indexes', 'Subtle_removed_entities_pylucene')

	load_path = os.path.join(ROOT_PATH, 'datasets', 'faqs_questions_pylucene_no_duplicates_16.11.txt')
	directory = os.path.join('indexes', 'FAQs_portuguese_analyser_16.11')

	start_time = time.time()
	print("Started creating the index")

	l_indexer(directory, load_path)

	print("Finished creating index successfully")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')

	# l = l_searcher("Olá?", directory, 10)

	


'''
Module used to extract semantic features from a given corpus.
'''

import os
import collections

import pandas as pd

from ASAPPpy.scripts.tools import preprocessing

def compute_ner(ners_corpus):
	""" Function used to compute the NERs difference between two sentences in the corpus """

	usable_ners = ['B-ABSTRACCAO', 'B-ACONTECIMENTO', 'B-COISA', 'B-LOCAL', 'B-OBRA', 'B-ORGANIZACAO', 'B-OUTRO', 'B-PESSOA', 'B-TEMPO', 'B-VALOR']

	ners_corpus_counts = []

	for i in range(0, len(ners_corpus), 2):
		ners_pair_counts = []
		total_ners_q = 0
		total_ners_r = 0

		for ner in usable_ners:
			temp_q = ners_corpus[i].count(ner)
			temp_r = ners_corpus[i+1].count(ner)

			difference = abs(temp_q-temp_r)
			ners_pair_counts.append(difference)

			total_ners_q += temp_q
			total_ners_r += temp_r

		all_ners_difference = abs(total_ners_q-total_ners_r)
		ners_pair_counts.insert(0, all_ners_difference)

		ners_corpus_counts.append(ners_pair_counts)

	usable_ners.insert(0, "all_ners")

	# convert list to Dataframe in order to use it with feature extraction
	ners_corpus_counts = pd.DataFrame(ners_corpus_counts)
	ners_corpus_counts.columns = usable_ners

	return ners_corpus_counts

def semantic_relations_coefficient(sentence1, sentence2, dictionaries_list):
	""" Auxiliar function used to compute the semantic relations coefficient between two sentences in the corpus """

	counters = len(dictionaries_list) * [0]
	normalization_counters = len(dictionaries_list) * [0]

	for count, relation in enumerate(dictionaries_list):
		for word in sentence1:
			if word in relation.keys():
				for definition in relation[word]:
					if definition in sentence2:
						counters[count] += 1
			if counters[count] != 0:
				normalization_counters[count] += 1

	normalized_counters = []

	for i in range(len(counters)):
		if counters[i] == 0 or normalization_counters[i] == 0:
			normalized_counters.append(0)
		else:
			normalized_counters.append(counters[i]/normalization_counters[i])

	# used to create a single feature for all semantic relations
	# if sum(normalized_counters) == 0:
	# 	relations_coefficient = 0
	# else:
	# 	relations_coefficient = sum(normalized_counters)/len(dictionaries_list)

	# return relations_coefficient

	# used to create a feature for each semantic relation
	if sum(normalized_counters[0:3]) == 0:
		antonym_feature = 0
	else:
		antonym_feature = sum(normalized_counters[0:3])/4

	if sum(normalized_counters[4:7]) == 0:
		synonym_feature = 0
	else:
		synonym_feature = sum(normalized_counters[4:7])/4

	if sum(normalized_counters[8:9]) == 0:
		hyperonym_feature = 0
	else:
		hyperonym_feature = sum(normalized_counters[8:9])/2

	other_feature = normalized_counters[10]

	return antonym_feature, synonym_feature, hyperonym_feature, other_feature

def compute_semantic_relations(pipe_lemmas, number_of_sources=3):
	""" Function used to compute the semantic relations coefficient between two sentences in the corpus """

	relations_file_path = os.path.join('semantic_relations', 'triplos_10recs', 'triplos_todos_10recs_n.txt')

	with open(relations_file_path) as relations_file:
		relations_data = relations_file.read().splitlines()

	relations_file.close()

	#there is a problem with rstrip, try to fix it in order to remove the tab
	relations_data_strip = [line.replace('\t', ' ') for line in relations_data]
	relations_data_split = [[relation] for relation in relations_data_strip]
	relations_data_string_split = [relation[0].split() for relation in relations_data_split]

	#keep only the semantic relations that were extracted from more than number_of_sources sources
	relations_sources = [relation for relation in relations_data_string_split if int(relation[3]) >= number_of_sources]

	antonym_strings = ["ANTONIMO_N_DE", "ANTONIMO_V_DE", "ANTONIMO_ADJ_DE", "ANTONIMO_ADV_DE"]
	synonym_strings = ["SINONIMO_N_DE", "SINONIMO_V_DE", "SINONIMO_ADJ_DE", "SINONIMO_ADV_DE"]
	hyperonym_strings = ["HIPERONIMO_DE", "HIPERONIMO_ACCAO_DE"]

	antonimo_n_de = collections.defaultdict(list)
	antonimo_v_de = collections.defaultdict(list)
	antonimo_adj_de = collections.defaultdict(list)
	antonimo_adv_de = collections.defaultdict(list)

	sinonimo_n_de = collections.defaultdict(list)
	sinonimo_v_de = collections.defaultdict(list)
	sinonimo_adj_de = collections.defaultdict(list)
	sinonimo_adv_de = collections.defaultdict(list)

	hiperonimo_de = collections.defaultdict(list)
	hiperonimo_accao_de = collections.defaultdict(list)

	outra = collections.defaultdict(list)

	for relation in relations_sources:
		if relation[1] == antonym_strings[0]:
			antonimo_n_de[relation[0]].append(relation[2])
		elif relation[1] == antonym_strings[1]:
			antonimo_v_de[relation[0]].append(relation[2])
		elif relation[1] == antonym_strings[2]:
			antonimo_adj_de[relation[0]].append(relation[2])
		elif relation[1] == antonym_strings[3]:
			antonimo_adv_de[relation[0]].append(relation[2])
		elif relation[1] == synonym_strings[0]:
			sinonimo_n_de[relation[0]].append(relation[2])
		elif relation[1] == synonym_strings[1]:
			sinonimo_v_de[relation[0]].append(relation[2])
		elif relation[1] == synonym_strings[2]:
			sinonimo_adj_de[relation[0]].append(relation[2])
		elif relation[1] == synonym_strings[3]:
			sinonimo_adv_de[relation[0]].append(relation[2])
		elif relation[1] == hyperonym_strings[0]:
			hiperonimo_de[relation[0]].append(relation[2])
		elif relation[1] == hyperonym_strings[1]:
			hiperonimo_accao_de[relation[0]].append(relation[2])
		else:
			outra[relation[0]].append(relation[2])

	dictionaries_relations_list = []

	dictionaries_relations_list.append(antonimo_n_de)
	dictionaries_relations_list.append(antonimo_v_de)
	dictionaries_relations_list.append(antonimo_adj_de)
	dictionaries_relations_list.append(antonimo_adv_de)
	dictionaries_relations_list.append(sinonimo_n_de)
	dictionaries_relations_list.append(sinonimo_v_de)
	dictionaries_relations_list.append(sinonimo_adj_de)
	dictionaries_relations_list.append(sinonimo_adv_de)
	dictionaries_relations_list.append(hiperonimo_de)
	dictionaries_relations_list.append(hiperonimo_accao_de)
	dictionaries_relations_list.append(outra)

	preprocessed_lemmatized_corpus = preprocessing(pipe_lemmas, 1, 1, 0, 0)

	# create only one semantic feature
	#preprocessed_lemmatized_corpus['semantics'] = preprocessed_lemmatized_corpus.apply(lambda pair: semantic_relations_coefficient(pair['text'], pair['response'], dictionaries_relations_list), axis=1)

	# create one semantic feature for each semantic relation
	preprocessed_lemmatized_corpus[['antonyms', 'synonyms', 'hyperonyms', 'other']] = preprocessed_lemmatized_corpus.apply(lambda pair: pd.Series(semantic_relations_coefficient(pair['text'], pair['response'], dictionaries_relations_list)), axis=1)

	return preprocessed_lemmatized_corpus



import sys
import feature_extraction as fe

from importlib import reload

word2vec_model = None
fasttext_model = None
ptlkb64_model = None
glove300_model = None
numberbatch_model = None

if __name__ == "__main__":
    models_loaded = 0
    while True:
        if models_loaded == 0:
            word2vec_model, fasttext_model, ptlkb64_model, glove300_model, numberbatch_model = fe.load_embeddings_models()
            models_loaded = 1
        fe.run_feature_extraction(word2vec_model=word2vec_model, fasttext_model=fasttext_model, ptlkb64_model=ptlkb64_model, glove300_model=glove300_model, numberbatch_model=numberbatch_model)
        print("Press enter to re-run the script, CTRL-C to exit")
        sys.stdin.readline()
        reload(fe)
        


'''
Module used for feature extraction of a corpus.
'''

import time
import os
import collections
import argparse
import spacy
import numpy as np
from gensim.models import KeyedVectors
from xml.etree import cElementTree as ET
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.svm import SVR
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.ensemble import BaggingRegressor
from sklearn.ensemble import AdaBoostRegressor
from sklearn.ensemble import VotingRegressor
from sklearn.model_selection import GridSearchCV
from joblib import dump, load
from NLPyPort.FullPipeline import new_full_pipe, load_congif_to_list
from scripts.xml_reader import read_xml
from scripts.xml_reader import read_xml_no_attributes

import models.word2vec.word2vec as w2c
import models.fastText.fasttext as ftt
import models.ontoPT.ptlkb as ptlkb
import load_embeddings as ld_emb
import scripts.tools as tl

from feature_engineering.lexical_features import create_word_ngrams, create_multiple_word_ngrams, create_character_ngrams, create_multiple_character_ngrams, compute_jaccard, compute_dice, compute_overlap, NG
from feature_engineering.syntactic_features import compute_pos, dependency_parsing
from feature_engineering.semantic_features import compute_ner, compute_semantic_relations
from feature_selection.feature_selection import feature_selection

config_list = load_congif_to_list()

def extract_features(corpus, preprocessed_corpus, word2vec_mdl=None, fasttext_mdl=None, ptlkb64_mdl=None, glove300_mdl=None, numberbatch_mdl=None):
	""" Function used to extract the features """

	# start_time = time.time()
	# print("Started running the pipeline")

	pipeline_output = new_full_pipe(corpus, options={"string_or_array":True}, config_list=config_list)

	# there's still a bug in the pipeline that makes it output two EOS tokens at the end of each run, reason why we read the output to the penultimate token.
	tags = tl.build_sentences_from_tokens(pipeline_output.pos_tags)
	lemmas = tl.build_sentences_from_tokens(pipeline_output.lemas)
	entities = tl.build_sentences_from_tokens(pipeline_output.entities)

	# print("Finished running the pipeline successfully")
	# print("--- %s seconds ---" %(time.time() - start_time))
	# print('\a')

	features = []

	# create word ngrams of different sizes
	word_ngrams_1 = create_word_ngrams(preprocessed_corpus, 1)

	word_ngrams_2 = create_word_ngrams(preprocessed_corpus, 2)

	word_ngrams_3 = create_word_ngrams(preprocessed_corpus, 3)

	# compute distance coefficients for these ngrams
	wn_jaccard_1 = compute_jaccard(word_ngrams_1)

	wn_jaccard_2 = compute_jaccard(word_ngrams_2)

	wn_jaccard_3 = compute_jaccard(word_ngrams_3)


	wn_dice_1 = compute_dice(word_ngrams_1)

	wn_dice_2 = compute_dice(word_ngrams_2)

	wn_dice_3 = compute_dice(word_ngrams_3)


	wn_overlap_1 = compute_overlap(word_ngrams_1)

	wn_overlap_2 = compute_overlap(word_ngrams_2)

	wn_overlap_3 = compute_overlap(word_ngrams_3)

	# create character ngrams of different sizes
	character_ngrams_2 = create_character_ngrams(preprocessed_corpus, 2)

	character_ngrams_3 = create_character_ngrams(preprocessed_corpus, 3)

	character_ngrams_4 = create_character_ngrams(preprocessed_corpus, 4)

	# compute distance coefficients for these ngrams
	cn_jaccard_2 = compute_jaccard(character_ngrams_2)

	cn_jaccard_3 = compute_jaccard(character_ngrams_3)

	cn_jaccard_4 = compute_jaccard(character_ngrams_4)


	cn_dice_2 = compute_dice(character_ngrams_2)

	cn_dice_3 = compute_dice(character_ngrams_3)

	cn_dice_4 = compute_dice(character_ngrams_4)


	cn_overlap_2 = compute_overlap(character_ngrams_2)

	cn_overlap_3 = compute_overlap(character_ngrams_3)

	cn_overlap_4 = compute_overlap(character_ngrams_4)

	if word2vec_mdl:
		word2vec = w2c.word2vec_model(word2vec_mdl, corpus, 0, 1, 0)

		word2vec_tfidf = w2c.word2vec_model(word2vec_mdl, corpus, 1, 1, 0)

	if fasttext_mdl:
		fasttext = ftt.fasttext_model(fasttext_mdl, corpus, 0, 1, 0)

		fasttext_tfidf = ftt.fasttext_model(fasttext_mdl, corpus, 1, 1, 0)

	if ptlkb64_mdl:
		ptlkb_64 = ptlkb.ptlkb_model(ptlkb64_mdl, 0, 1, 0, lemmas)

		ptlkb_64_tfidf = ptlkb.ptlkb_model(ptlkb64_mdl, 1, 1, 0, lemmas)

	if glove300_mdl:
		glove_300 = ld_emb.word_embeddings_model(glove300_mdl, corpus, 0, 1, 0)

		glove_300_tfidf = ld_emb.word_embeddings_model(glove300_mdl, corpus, 1, 1, 0)

	# compute tfidf matrix - padding was applied to vectors of different sizes by adding zeros to the smaller vector of the pair
	tfidf_corpus = tl.preprocessing(corpus, 0, 0, 0, 1)
	tfidf_matrix = tl.compute_tfidf_matrix(tfidf_corpus, 0, 0, 1)

	# compute semantic relations coefficients
	semantic_relations = compute_semantic_relations(lemmas)

	# compute POS tags
	pos_tags = compute_pos(tags)

	# compute NERs
	ners = compute_ner(entities)

	# compute Syntactic Dependency parsing
	dependencies = dependency_parsing(corpus)

	# create multiple word ngrams of different sizes
	word_ngrams_1_2 = create_multiple_word_ngrams(preprocessed_corpus, 1, 2)

	# compute the cosine similarity between the multiple word ngrams converted sentences
	wn_cosine_1_2 = NG(word_ngrams_1_2)

	# create multiple character ngrams of different sizes
	character_ngrams_1_2_3 = create_multiple_character_ngrams(preprocessed_corpus, 1, 2, 3)

	# compute the cosine similarity between the multiple character ngrams converted sentences
	cn_cosine_1_2_3 = NG(character_ngrams_1_2_3)

	if numberbatch_mdl:
		numberbatch = ld_emb.word_embeddings_model(numberbatch_mdl, corpus, 0, 1, 0)

		numberbatch_tfidf = ld_emb.word_embeddings_model(numberbatch_mdl, corpus, 1, 1, 0)

	for pair in range(len(preprocessed_corpus)):
		flag = 1
		features_pair = []
		used_features = [False] * 66

		if flag == 1:
			used_features[0] = True
			features_pair.append(wn_jaccard_1['jaccard'][pair])
		if flag == 1:
			used_features[1] = True
			features_pair.append(wn_jaccard_2['jaccard'][pair])
		if flag == 1:
			used_features[2] = True
			features_pair.append(wn_jaccard_3['jaccard'][pair])
		if flag == 1:
			used_features[3] = True
			features_pair.append(wn_dice_1['dice'][pair])
		if flag == 1:
			used_features[4] = True
			features_pair.append(wn_dice_2['dice'][pair])
		if flag == 1:
			used_features[5] = True
			features_pair.append(wn_dice_3['dice'][pair])
		if flag == 1:
			used_features[6] = True
			features_pair.append(wn_overlap_1['overlap'][pair])
		if flag == 1:
			used_features[7] = True
			features_pair.append(wn_overlap_2['overlap'][pair])
		if flag == 1:
			used_features[8] = True
			features_pair.append(wn_overlap_3['overlap'][pair])
		if flag == 1:
			used_features[9] = True
			features_pair.append(cn_jaccard_2['jaccard'][pair])
		if flag == 1:
			used_features[10] = True
			features_pair.append(cn_jaccard_3['jaccard'][pair])
		if flag == 1:
			used_features[11] = True
			features_pair.append(cn_jaccard_4['jaccard'][pair])
		if flag == 1:
			used_features[12] = True
			features_pair.append(cn_dice_2['dice'][pair])
		if flag == 1:
			used_features[13] = True
			features_pair.append(cn_dice_3['dice'][pair])
		if flag == 1:
			used_features[14] = True
			features_pair.append(cn_dice_4['dice'][pair])
		if flag == 1:
			used_features[15] = True
			features_pair.append(cn_overlap_2['overlap'][pair])
		if flag == 1:
			used_features[16] = True
			features_pair.append(cn_overlap_3['overlap'][pair])
		if flag == 1:
			used_features[17] = True
			features_pair.append(cn_overlap_4['overlap'][pair])
		if flag == 1:
			used_features[18] = True
			features_pair.append(word2vec[pair])
		if flag == 1:
			used_features[19] = True
			features_pair.append(word2vec_tfidf[pair])
		if flag == 1:
			used_features[20] = True
			features_pair.append(fasttext[pair])
		if flag == 1:
			used_features[21] = True
			features_pair.append(fasttext_tfidf[pair])
		if flag == 1:
			used_features[22] = True
			features_pair.append(ptlkb_64[pair])
		if flag == 1:
			used_features[23] = True
			features_pair.append(ptlkb_64_tfidf[pair])
		if flag == 1:
			used_features[24] = True
			features_pair.append(glove_300[pair])
		if flag == 1:
			used_features[25] = True
			features_pair.append(glove_300_tfidf[pair])
		if flag == 1:
			used_features[26] = True
			features_pair.append(tfidf_matrix[pair])
		if flag == 1:
			used_features[27] = True
			features_pair.append(semantic_relations['antonyms'][pair])
		if flag == 1:
			used_features[28] = True
			features_pair.append(semantic_relations['synonyms'][pair])
		if flag == 1:
			used_features[29] = True
			features_pair.append(semantic_relations['hyperonyms'][pair])
		if flag == 1:
			used_features[30] = True
			features_pair.append(semantic_relations['other'][pair])
		if flag == 1 and ('n' in pos_tags.columns):
			used_features[31] = True
			features_pair.append(pos_tags['n'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('prop' in pos_tags.columns):
			used_features[32] = True
			features_pair.append(pos_tags['prop'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('adj' in pos_tags.columns):
			used_features[33] = True
			features_pair.append(pos_tags['adj'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('n-adj' in pos_tags.columns):
			used_features[34] = True
			features_pair.append(pos_tags['n-adj'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('v-fin' in pos_tags.columns):
			used_features[35] = True
			features_pair.append(pos_tags['v-fin'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('v-inf' in pos_tags.columns):
			used_features[36] = True
			features_pair.append(pos_tags['v-inf'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('v-pcp' in pos_tags.columns):
			used_features[37] = True
			features_pair.append(pos_tags['v-pcp'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('v-ger' in pos_tags.columns):
			used_features[38] = True
			features_pair.append(pos_tags['v-ger'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('art' in pos_tags.columns):
			used_features[39] = True
			features_pair.append(pos_tags['art'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('pron-pers' in pos_tags.columns):
			used_features[40] = True
			features_pair.append(pos_tags['pron-pers'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('pron-det' in pos_tags.columns):
			used_features[41] = True
			features_pair.append(pos_tags['pron-det'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('pron-indp' in pos_tags.columns):
			used_features[42] = True
			features_pair.append(pos_tags['pron-indp'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('adv' in pos_tags.columns):
			used_features[43] = True
			features_pair.append(pos_tags['adv'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('num' in pos_tags.columns):
			used_features[44] = True
			features_pair.append(pos_tags['num'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('prp' in pos_tags.columns):
			used_features[45] = True
			features_pair.append(pos_tags['prp'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('intj' in pos_tags.columns):
			used_features[46] = True
			features_pair.append(pos_tags['intj'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('conj-s' in pos_tags.columns):
			used_features[47] = True
			features_pair.append(pos_tags['conj-s'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('conj-c' in pos_tags.columns):
			used_features[48] = True
			features_pair.append(pos_tags['conj-c'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('punc' in pos_tags.columns):
			used_features[49] = True
			features_pair.append(pos_tags['punc'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('all_ners' in ners.columns):
			used_features[50] = True
			features_pair.append(ners['all_ners'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('B-ABSTRACCAO' in ners.columns):
			used_features[51] = True
			features_pair.append(ners['B-ABSTRACCAO'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('B-ACONTECIMENTO' in ners.columns):
			used_features[52] = True
			features_pair.append(ners['B-ACONTECIMENTO'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('B-COISA' in ners.columns):
			used_features[53] = True
			features_pair.append(ners['B-COISA'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('B-LOCAL' in ners.columns):
			used_features[54] = True
			features_pair.append(ners['B-LOCAL'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('B-OBRA' in ners.columns):
			used_features[55] = True
			features_pair.append(ners['B-OBRA'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('B-ORGANIZACAO' in ners.columns):
			used_features[56] = True
			features_pair.append(ners['B-ORGANIZACAO'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('B-OUTRO' in ners.columns):
			used_features[57] = True
			features_pair.append(ners['B-OUTRO'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('B-PESSOA' in ners.columns):
			used_features[58] = True
			features_pair.append(ners['B-PESSOA'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('B-TEMPO' in ners.columns):
			used_features[59] = True
			features_pair.append(ners['B-TEMPO'][pair])
		else:
			features_pair.append(0)
		if flag == 1 and ('B-VALOR' in ners.columns):
			used_features[60] = True
			features_pair.append(ners['B-VALOR'][pair])
		else:
			features_pair.append(0)
		if flag == 1:
			used_features[61] = True
			features_pair.append(dependencies['dependency_parsing_jc'][pair])
		if flag == 1:
			used_features[62] = True
			features_pair.append(wn_cosine_1_2['NG'][pair])
		if flag == 1:
			used_features[63] = True
			features_pair.append(cn_cosine_1_2_3['NG'][pair])
		if flag == 1:
			used_features[64] = True
			features_pair.append(numberbatch[pair])
		if flag == 1:
			used_features[65] = True
			features_pair.append(numberbatch_tfidf[pair])

		tuple_features_pair = tuple(features_pair)
		features.append(tuple_features_pair)

	return np.array(features), used_features

def load_features(f_selection, corpus, preprocessed_corpus, word2vec_mdl=None, fasttext_mdl=None, ptlkb64_mdl=None, glove300_mdl=None, numberbatch_mdl=None):
	""" Function used to extract the features """

	# start_time = time.time()
	# print("Started running the pipeline")

	pipeline_output = new_full_pipe(corpus, options={"string_or_array":True}, config_list=config_list)

	# there's still a bug in the pipeline that makes it output two EOS tokens at the end of each run, reason why we read the output to the penultimate token.
	tags = tl.build_sentences_from_tokens(pipeline_output.pos_tags)
	lemmas = tl.build_sentences_from_tokens(pipeline_output.lemas)
	entities = tl.build_sentences_from_tokens(pipeline_output.entities)

	# print("Finished running the pipeline successfully")
	# print("--- %s seconds ---" %(time.time() - start_time))
	# print('\a')

	features = []

	# create word ngrams of different sizes
	if 1 in f_selection[0:9:3]:
		word_ngrams_1 = create_word_ngrams(preprocessed_corpus, 1)

	if 1 in f_selection[1:9:3]:
		word_ngrams_2 = create_word_ngrams(preprocessed_corpus, 2)

	if 1 in f_selection[2:9:3]:
		word_ngrams_3 = create_word_ngrams(preprocessed_corpus, 3)

	# compute distance coefficients for these ngrams
	if f_selection[0]:
		wn_jaccard_1 = compute_jaccard(word_ngrams_1)
	if f_selection[1]:
		wn_jaccard_2 = compute_jaccard(word_ngrams_2)
	if f_selection[2]:
		wn_jaccard_3 = compute_jaccard(word_ngrams_3)

	if f_selection[3]:
		wn_dice_1 = compute_dice(word_ngrams_1)
	if f_selection[4]:
		wn_dice_2 = compute_dice(word_ngrams_2)
	if f_selection[5]:
		wn_dice_3 = compute_dice(word_ngrams_3)

	if f_selection[6]:
		wn_overlap_1 = compute_overlap(word_ngrams_1)
	if f_selection[7]:
		wn_overlap_2 = compute_overlap(word_ngrams_2)
	if f_selection[8]:
		wn_overlap_3 = compute_overlap(word_ngrams_3)

	# create character ngrams of different sizes
	if 1 in f_selection[9:18:3]:
		character_ngrams_2 = create_character_ngrams(preprocessed_corpus, 2)
	if 1 in f_selection[10:18:3]:
		character_ngrams_3 = create_character_ngrams(preprocessed_corpus, 3)
	if 1 in f_selection[11:18:3]:
		character_ngrams_4 = create_character_ngrams(preprocessed_corpus, 4)

	# compute distance coefficients for these ngrams
	if f_selection[9]:
		cn_jaccard_2 = compute_jaccard(character_ngrams_2)
	if f_selection[10]:
		cn_jaccard_3 = compute_jaccard(character_ngrams_3)
	if f_selection[11]:
		cn_jaccard_4 = compute_jaccard(character_ngrams_4)

	if f_selection[12]:
		cn_dice_2 = compute_dice(character_ngrams_2)
	if f_selection[13]:
		cn_dice_3 = compute_dice(character_ngrams_3)
	if f_selection[14]:
		cn_dice_4 = compute_dice(character_ngrams_4)

	if f_selection[15]:
		cn_overlap_2 = compute_overlap(character_ngrams_2)
	if f_selection[16]:
		cn_overlap_3 = compute_overlap(character_ngrams_3)
	if f_selection[17]:
		cn_overlap_4 = compute_overlap(character_ngrams_4)

	if word2vec_mdl:
		if f_selection[18]:
			word2vec = w2c.word2vec_model(word2vec_mdl, corpus, 0, 1, 0)
		if f_selection[19]:
			word2vec_tfidf = w2c.word2vec_model(word2vec_mdl, corpus, 1, 1, 0)

	if fasttext_mdl:
		if f_selection[20]:
			fasttext = ftt.fasttext_model(fasttext_mdl, corpus, 0, 1, 0)
		if f_selection[21]:
			fasttext_tfidf = ftt.fasttext_model(fasttext_mdl, corpus, 1, 1, 0)

	if ptlkb64_mdl:
		if f_selection[22]:
			ptlkb_64 = ptlkb.ptlkb_model(ptlkb64_mdl, 0, 1, 0, lemmas)
		if f_selection[23]:
			ptlkb_64_tfidf = ptlkb.ptlkb_model(ptlkb64_mdl, 1, 1, 0, lemmas)

	if glove300_mdl:
		if f_selection[24]:
			glove_300 = ld_emb.word_embeddings_model(glove300_mdl, corpus, 0, 1, 0)
		if f_selection[25]:
			glove_300_tfidf = ld_emb.word_embeddings_model(glove300_mdl, corpus, 1, 1, 0)

	# compute tfidf matrix - padding was applied to vectors of different sizes by adding zeros to the smaller vector of the pair
	if f_selection[26]:
		tfidf_corpus = tl.preprocessing(corpus, 0, 0, 0, 1)
		tfidf_matrix = tl.compute_tfidf_matrix(tfidf_corpus, 0, 0, 1)

	# compute semantic relations coefficients
	if 1 in f_selection[27:31]:
		semantic_relations = compute_semantic_relations(lemmas)

	# compute POS tags
	if 1 in f_selection[31:50]:
		pos_tags = compute_pos(tags)

	# compute NERs
	if 1 in f_selection[50:61]:
		ners = compute_ner(entities)

	# compute Syntactic Dependency parsing
	if f_selection[61]:
		dependencies = dependency_parsing(corpus)

	# create multiple word ngrams of different sizes
	if f_selection[62]:
		word_ngrams_1_2 = create_multiple_word_ngrams(preprocessed_corpus, 1, 2)

	# compute the cosine similarity between the multiple word ngrams converted sentences
	if f_selection[62]:
		wn_cosine_1_2 = NG(word_ngrams_1_2)

	# create multiple character ngrams of different sizes
	if f_selection[63]:
		character_ngrams_1_2_3 = create_multiple_character_ngrams(preprocessed_corpus, 1, 2, 3)

	# compute the cosine similarity between the multiple character ngrams converted sentences
	if f_selection[63]:
		cn_cosine_1_2_3 = NG(character_ngrams_1_2_3)

	if numberbatch_mdl:
		if f_selection[64]:
			numberbatch = ld_emb.word_embeddings_model(numberbatch_mdl, corpus, 0, 1, 0)
		if f_selection[65]:
			numberbatch_tfidf = ld_emb.word_embeddings_model(numberbatch_mdl, corpus, 1, 1, 0)

	for pair in range(len(preprocessed_corpus)):
		if f_selection is not None:
			features_pair = []

			if f_selection[0]:
				features_pair.append(wn_jaccard_1['jaccard'][pair])
			if f_selection[1]:
				features_pair.append(wn_jaccard_2['jaccard'][pair])
			if f_selection[2]:
				features_pair.append(wn_jaccard_3['jaccard'][pair])
			if f_selection[3]:
				features_pair.append(wn_dice_1['dice'][pair])
			if f_selection[4]:
				features_pair.append(wn_dice_2['dice'][pair])
			if f_selection[5]:
				features_pair.append(wn_dice_3['dice'][pair])
			if f_selection[6]:
				features_pair.append(wn_overlap_1['overlap'][pair])
			if f_selection[7]:
				features_pair.append(wn_overlap_2['overlap'][pair])
			if f_selection[8]:
				features_pair.append(wn_overlap_3['overlap'][pair])
			if f_selection[9]:
				features_pair.append(cn_jaccard_2['jaccard'][pair])
			if f_selection[10]:
				features_pair.append(cn_jaccard_3['jaccard'][pair])
			if f_selection[11]:
				features_pair.append(cn_jaccard_4['jaccard'][pair])
			if f_selection[12]:
				features_pair.append(cn_dice_2['dice'][pair])
			if f_selection[13]:
				features_pair.append(cn_dice_3['dice'][pair])
			if f_selection[14]:
				features_pair.append(cn_dice_4['dice'][pair])
			if f_selection[15]:
				features_pair.append(cn_overlap_2['overlap'][pair])
			if f_selection[16]:
				features_pair.append(cn_overlap_3['overlap'][pair])
			if f_selection[17]:
				features_pair.append(cn_overlap_4['overlap'][pair])
			if f_selection[18]:
				features_pair.append(word2vec[pair])
			if f_selection[19]:
				features_pair.append(word2vec_tfidf[pair])
			if f_selection[20]:
				features_pair.append(fasttext[pair])
			if f_selection[21]:
				features_pair.append(fasttext_tfidf[pair])
			if f_selection[22]:
				features_pair.append(ptlkb_64[pair])
			if f_selection[23]:
				features_pair.append(ptlkb_64_tfidf[pair])
			if f_selection[24]:
				features_pair.append(glove_300[pair])
			if f_selection[25]:
				features_pair.append(glove_300_tfidf[pair])
			if f_selection[26]:
				features_pair.append(tfidf_matrix[pair])
			if f_selection[27]:
				features_pair.append(semantic_relations['antonyms'][pair])
			if f_selection[28]:
				features_pair.append(semantic_relations['synonyms'][pair])
			if f_selection[29]:
				features_pair.append(semantic_relations['hyperonyms'][pair])
			if f_selection[30]:
				features_pair.append(semantic_relations['other'][pair])
			if f_selection[31] and ('n' in pos_tags.columns):
				features_pair.append(pos_tags['n'][pair])
			if f_selection[32] and ('prop' in pos_tags.columns):
				features_pair.append(pos_tags['prop'][pair])
			if f_selection[33] and ('adj' in pos_tags.columns):
				features_pair.append(pos_tags['adj'][pair])
			if f_selection[34] and ('n-adj' in pos_tags.columns):
				features_pair.append(pos_tags['n-adj'][pair])
			if f_selection[35] and ('v-fin' in pos_tags.columns):
				features_pair.append(pos_tags['v-fin'][pair])
			if f_selection[36] and ('v-inf' in pos_tags.columns):
				features_pair.append(pos_tags['v-inf'][pair])
			if f_selection[37] and ('v-pcp' in pos_tags.columns):
				features_pair.append(pos_tags['v-pcp'][pair])
			if f_selection[38] and ('v-ger' in pos_tags.columns):
				features_pair.append(pos_tags['v-ger'][pair])
			if f_selection[39] and ('art' in pos_tags.columns):
				features_pair.append(pos_tags['art'][pair])
			if f_selection[40] and ('pron-pers' in pos_tags.columns):
				features_pair.append(pos_tags['pron-pers'][pair])
			if f_selection[41] and ('pron-det' in pos_tags.columns):
				features_pair.append(pos_tags['pron-det'][pair])
			if f_selection[42] and ('pron-indp' in pos_tags.columns):
				features_pair.append(pos_tags['pron-indp'][pair])
			if f_selection[43] and ('adv' in pos_tags.columns):
				features_pair.append(pos_tags['adv'][pair])
			if f_selection[44] and ('num' in pos_tags.columns):
				features_pair.append(pos_tags['num'][pair])
			if f_selection[45] and ('prp' in pos_tags.columns):
				features_pair.append(pos_tags['prp'][pair])
			if f_selection[46] and ('intj' in pos_tags.columns):
				features_pair.append(pos_tags['intj'][pair])
			if f_selection[47] and ('conj-s' in pos_tags.columns):
				features_pair.append(pos_tags['conj-s'][pair])
			if f_selection[48] and ('conj-c' in pos_tags.columns):
				features_pair.append(pos_tags['conj-c'][pair])
			if f_selection[49] and ('punc' in pos_tags.columns):
				features_pair.append(pos_tags['punc'][pair])
			if f_selection[50] and ('all_ners' in ners.columns):
				features_pair.append(ners['all_ners'][pair])
			if f_selection[51] and ('B-ABSTRACCAO' in ners.columns):
				features_pair.append(ners['B-ABSTRACCAO'][pair])
			if f_selection[52] and ('B-ACONTECIMENTO' in ners.columns):
				features_pair.append(ners['B-ACONTECIMENTO'][pair])
			if f_selection[53] and ('B-COISA' in ners.columns):
				features_pair.append(ners['B-COISA'][pair])
			if f_selection[54] and ('B-LOCAL' in ners.columns):
				features_pair.append(ners['B-LOCAL'][pair])
			if f_selection[55] and ('B-OBRA' in ners.columns):
				features_pair.append(ners['B-OBRA'][pair])
			if f_selection[56] and ('B-ORGANIZACAO' in ners.columns):
				features_pair.append(ners['B-ORGANIZACAO'][pair])
			if f_selection[57] and ('B-OUTRO' in ners.columns):
				features_pair.append(ners['B-OUTRO'][pair])
			if f_selection[58] and ('B-PESSOA' in ners.columns):
				features_pair.append(ners['B-PESSOA'][pair])
			if f_selection[59] and ('B-TEMPO' in ners.columns):
				features_pair.append(ners['B-TEMPO'][pair])
			if f_selection[60] and ('B-VALOR' in ners.columns):
				features_pair.append(ners['B-VALOR'][pair])
			if f_selection[61]:
				features_pair.append(dependencies['dependency_parsing_jc'][pair])
			if f_selection[62]:
				features_pair.append(wn_cosine_1_2['NG'][pair])
			if f_selection[63]:
				features_pair.append(cn_cosine_1_2_3['NG'][pair])
			if f_selection[64]:
				features_pair.append(numberbatch[pair])
			if f_selection[65]:
				features_pair.append(numberbatch_tfidf[pair])

			tuple_features_pair = tuple(features_pair)
			features.append(tuple_features_pair)

	return np.array(features)

def load_embeddings_models():
	""" Function used to load the word-embedding models """

	# ---LOADING WORD2VEC MODEL---
	model_load_path = os.path.join('models', 'word2vec', 'NILC', 'nilc_cbow_s300_300k.txt')
	start_time = time.time()
	print("Started loading the word2vec model")
	word2vec_model = KeyedVectors.load_word2vec_format(model_load_path)
	# word2vec_model = None
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')

	# ---LOADING FASTTEXT MODEL---
	model_path = os.path.join('models', 'fastText', 'cc.pt.300_300k.vec')
	start_time = time.time()
	print("Started loading the fasttext model")
	fasttext_model = KeyedVectors.load_word2vec_format(model_path)
	# fasttext_model = None
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')	

	# ---LOADING PT-LKB MODEL---
	model_load_path = os.path.join('models', 'ontoPT', 'PT-LKB_embeddings_64', 'ptlkb_64_30_200_p_str.emb')
	# model_load_path = os.path.join('models', 'ontoPT', 'PT-LKB_embeddings_128', 'ptlkb_128_80_10_p_str.emb')
	start_time = time.time()
	print("Started loading the PT-LKB-64 model")
	ptlkb64_model = KeyedVectors.load_word2vec_format(model_load_path)
	# ptlkb64_model = None
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')

	# ---LOADING GLOVE-300 MODEL---
	model_load_path = os.path.join('models', 'glove', 'glove_s300_300k.txt')
	start_time = time.time()
	print("Started loading the GLOVE 300 dimensions model")
	glove300_model = KeyedVectors.load_word2vec_format(model_load_path)
	# glove300_model = None
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')

	# ---LOADING NUMBERBATCH MODEL---
	model_load_path = os.path.join('models', 'numberbatch', 'numberbatch-17.02_pt_tratado.txt')
	start_time = time.time()
	print("Started loading the NUMBERBATCH dimensions model")
	numberbatch_model = KeyedVectors.load_word2vec_format(model_load_path)
	# numberbatch_model = None
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')

	return word2vec_model, fasttext_model, ptlkb64_model, glove300_model, numberbatch_model

def run_feature_extraction(word2vec_model=None, fasttext_model=None, ptlkb64_model=None, glove300_model=None, numberbatch_model=None):
	""" Function used to compute the models """

	parser = argparse.ArgumentParser(description=__doc__)
	parser.add_argument('test', help='XML file with test data')
	parser.add_argument('output', help='Output tagged XML file')
	args = parser.parse_args()

	"""
	system_mode = 0 -> uses the variant questions with the system
	system_mode = 1 -> uses the PTPT and PTBR train ASSIN collection datasets with the system
	system_mode = 2 -> uses the PTPT and PTBR train and test ASSIN collection datasets with the system
	system_mode = 3 -> uses the Whoosh collection with the system
	system_mode = 4 -> uses ASSIN 1 and ASSIN 2 training collection datasets with the system
	system_mode = 5 -> uses ASSIN 1 training and testing collection and ASSIN 2 training collection datasets with the system
	"""

	system_mode = 4

	# Flag to indicate if the extracted features should be written to a file (1) or not (0)
	features_to_file_flag = 0

	# extract labels
	train_pairs = []
	# train_pairs.extend(read_xml("datasets/assin/assin1/assin-ptpt-train.xml", need_labels=True))
	# train_pairs.extend(read_xml("datasets/assin/assin1/assin-ptbr-train.xml", need_labels=True))

	if system_mode == 2 or system_mode == 5:
		train_pairs.extend(read_xml("datasets/assin/assin1/assin-ptpt-test.xml", need_labels=True))
		train_pairs.extend(read_xml("datasets/assin/assin1/assin-ptbr-test.xml", need_labels=True))
	if system_mode == 4 or system_mode == 5:
		train_pairs.extend(read_xml("datasets/assin/assin2/assin2-train-only.xml", need_labels=True))

	train_similarity_target = np.array([pair.similarity for pair in train_pairs])

	# extract training features
	train_corpus = tl.read_corpus(train_pairs)

	# tl.write_data_to_file(train_corpus, "finetune.train.raw")
	# print("Wrote training corpus")

	# preprocessing(text, tokenization=0, rm_stopwords=0, numbers_to_text=0, to_tfidf=0)
	preprocessed_train_corpus = tl.preprocessing(train_corpus, 0, 0, 0, 0)
	train_features, used_train_features = extract_features(train_corpus, preprocessed_train_corpus, word2vec_mdl=word2vec_model, fasttext_mdl=fasttext_model, ptlkb64_mdl=ptlkb64_model, glove300_mdl=glove300_model, numberbatch_mdl=numberbatch_model)

	# write train features to a .csv file
	if features_to_file_flag == 1:
		tl.write_features_to_csv(train_pairs, train_features, "assin1-train-test-assin2-train-ftrain.csv")

	#############################################################
	test_pairs_dev = read_xml('datasets/assin/assin2/assin2-dev.xml', need_labels=False)

	test_corpus_dev = tl.read_corpus(test_pairs_dev)
	# preprocessing(text, tokenization=0, rm_stopwords=0, numbers_to_text=0, to_tfidf=0)
	preprocessed_test_corpus_dev = tl.preprocessing(test_corpus_dev, 0, 0, 0, 0)
	test_features_dev, used_test_features_dev = extract_features(test_corpus_dev, preprocessed_test_corpus_dev, word2vec_mdl=word2vec_model, fasttext_mdl=fasttext_model, ptlkb64_mdl=ptlkb64_model, glove300_mdl=glove300_model, numberbatch_mdl=numberbatch_model)

	test_pairs_selection = read_xml('datasets/assin/assin2/assin2-dev.xml', need_labels=True)
	test_similarity_target = np.array([pair.similarity for pair in test_pairs_selection])
	#############################################################

	# extract test features
	# test_pairs = read_xml(args.test, need_labels=False)

	# uncomment next line and comment previous one to compute ASSIN 2 submission results
	test_pairs = read_xml_no_attributes(args.test)

	test_corpus = tl.read_corpus(test_pairs)
	# preprocessing(text, tokenization=0, rm_stopwords=0, numbers_to_text=0, to_tfidf=0)
	preprocessed_test_corpus = tl.preprocessing(test_corpus, 0, 0, 0, 0)
	test_features, used_test_features = extract_features(test_corpus, preprocessed_test_corpus, word2vec_mdl=word2vec_model, fasttext_mdl=fasttext_model, ptlkb64_mdl=ptlkb64_model, glove300_mdl=glove300_model, numberbatch_mdl=numberbatch_model)

	# write test features to a .csv file
	if features_to_file_flag == 1:
		tl.write_features_to_csv(test_pairs, test_features, "assin1-train-test-assin2-train-ftest.csv")

	# extract test features for feature selection (labels needed in order to perform evaluation)
	# test_pairs_selection = read_xml(args.test, need_labels=True)
	# test_similarity_target = np.array([pair.similarity for pair in test_pairs_selection])

	'''
	In order to create a STS model, it is required to select a regressor algorithm. We rely on the scikit-learn implementation of these algorithms to run our system. The default regresso is SVR, however an algorithm of your choice can be used. Select one type of regressor from scikit-learn. Here is a list with some examples: 
		- GaussianProcessRegressor()
		- DecisionTreeRegressor()
		- LinearRegression()
		- BaggingRegressor(n_estimators=100)
		- AdaBoostRegressor(n_estimators=100)
		- GradientBoostingRegressor()
		- RandomForestRegressor(n_estimators=100)
	'''

	regressor = SVR(gamma='scale', C=10.0, kernel='rbf')

	# ensemble = VotingRegressor(estimators=[('svr', regressor_1), ('gb', regressor_2), ('rf', regressor_3)])

	# params = {'svr__C': [1.0, 10.0, 100.0], 'svr__kernel': ['linear', 'poly', 'rbf', 'sigmoid'], 'rf__n_estimators': [10, 20, 100, 200]}

	# params = {'kernel':('linear', 'poly', 'rbf', 'sigmoid')}

	# regressor = GridSearchCV(regressor_1, params, cv=5)

	use_feature_selection = 1

	if use_feature_selection:
		selected_selector, selected_train_features, selected_test_features = feature_selection(train_features, test_features_dev, train_similarity_target, test_similarity_target, regressor, used_train_features)

		test_features_selected = selected_selector.transform(test_features)

		regressor.fit(selected_train_features, train_similarity_target)

		# save model to disk
		model_save_path = os.path.join('trained_models', 'SVR_FS.joblib')
		dump(regressor, model_save_path)

		# apply model to the test dataset
		## this needs to be fixed in order to take advantage of the manual feature selection
		predicted_similarity = regressor.predict(test_features_selected)
	else:
		regressor.fit(train_features, train_similarity_target)

		# save model to disk
		model_save_path = os.path.join('trained_models', 'SVR_NFS.joblib')
		dump(regressor, model_save_path)

		# apply model to the test dataset
		predicted_similarity = regressor.predict(test_features)

	# write output
	tree = ET.parse(args.test)
	root = tree.getroot()
	for i in range(len(test_pairs)):
		pairs = root[i]
		pairs.set('entailment', "None")
		pairs.set('similarity', str(predicted_similarity[i]))

	tree.write(args.output, 'utf-8')

# if __name__ == '__main__':
# 	system_mode = 4
# 	run_pipeline = 1

# 	# Flag to indicate if the extracted features should be written to a file (1) or not (0)
# 	features_to_file_flag = 0

# 	# extract labels
# 	train_pairs = []
# 	# train_pairs.extend(read_xml("datasets/assin/assin1/assin-ptpt-train.xml", need_labels=True))
# 	# train_pairs.extend(read_xml("datasets/assin/assin1/assin-ptbr-train.xml", need_labels=True))

# 	if system_mode == 2 or system_mode == 5:
# 		# train_pairs.extend(read_xml("assin-ptpt-test.xml", need_labels=True))
# 		train_pairs.extend(read_xml("datasets/assin/assin1/assin-ptbr-test.xml", need_labels=True))
# 	if system_mode == 4 or system_mode == 5:
# 		train_pairs.extend(read_xml("datasets/assin/assin2/assin2-train-only.xml", need_labels=True))

# 	train_similarity_target = np.array([pair.similarity for pair in train_pairs])

# 	# extract training features
# 	train_corpus = tl.read_corpus(train_pairs)

# 	tl.write_data_to_file(train_corpus, "finetune.train.raw")
# 	print("Wrote training corpus")

	# test_pairs = read_xml_no_attributes("datasets/assin/assin2/assin2-blind-test.xml")

	# test_corpus = tl.read_corpus(test_pairs)

	# tl.write_data_to_file(test_corpus, "finetune.test.raw")
	# print("Wrote testing corpus")



import keras as k
import pandas as pd
import cufflinks as cf
from sklearn.model_selection import train_test_split
from keras import Sequential
from keras.layers import  Embedding
from keras.layers import Dense,SpatialDropout1D
from keras.preprocessing.text import Tokenizer
from keras.preprocessing.sequence import pad_sequences
from keras.layers.recurrent import LSTM
import numpy as np 
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from keras.preprocessing.text import Tokenizer
from keras.preprocessing.sequence import pad_sequences
from keras.models import Sequential
from keras.layers import Dense, Embedding, LSTM, SpatialDropout1D
from sklearn.model_selection import train_test_split
from keras.utils.np_utils import to_categorical
from keras.callbacks import EarlyStopping
from keras.layers import Dropout
from keras.models import Model, load_model
from sklearn.utils import resample
import tensorflow as tf
from keras.optimizers import Adam, RMSprop
from imblearn.over_sampling import SMOTE
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.naive_bayes import MultinomialNB
from sklearn.metrics import classification_report
from sklearn.naive_bayes import GaussianNB
import pickle
from sklearn import svm
from sklearn.feature_extraction.text import TfidfVectorizer
from ASAPPpy.classifiers.svm_binaria import *
#from svm_binaria import *
import matplotlib.pyplot as plt

#from sklearn.svm._classes import *
#naive bayes, svm, random forest



#https://github.com/susanli2016/NLP-with-Python/blob/master/Multi-Class%20Text%20Classification%20LSTM%20Consumer%20complaints.ipynb

#1---Empresa na Hora
#2---Marca na hora e marca na hora online
#3---Formas jurídicas
#4---Cartão de Empresa/Cartão de Pessoa Coletiva
#5---Criação da Empresa Online
#6---Certificados de Admissibilidade
#7---Inscrições Online
#8---Certidões Online
#9---Gestão da Empresa Online
#10---RJACSR
#11---Alojamento Local

def escala(X,y):
	from imblearn.over_sampling import RandomOverSampler
	sampler = RandomOverSampler(sampling_strategy='not majority',random_state=0)
	X_train, Y_train = sampler.fit_sample(X, y)
	return X_train,Y_train

def treina(model_name):

	df = pd.read_csv("input/multiclass_only_train.txt",sep='§',header=0)
	df.info()

	max_len = 0
	for value in df.Perguntas:
		if(len(value)>max_len):
			#print(value)
			max_len = len(value)
	max_words = 0
	for value in df.Perguntas:
		word_count = len(value.split(" "))
		if(word_count>max_words):
			#print(word_count)
			max_words = word_count
	#print("---------")
	#print(max_words)
	#print(df.Class.value_counts().sort_values())
	g = df.groupby('Class')
	g= pd.DataFrame(g.apply(lambda x: x.sample(g.size().min()).reset_index(drop=True)))
	#print(g)

	#df =  pd.DataFrame(g.apply(lambda x: x.sample(g.size().min()).reset_index(drop=True)))
	#print("###")
	#print(df.shape)
	#print(df.head(10))
	#df = g



	#X_conveted = pd.get_dummies(df["Perguntas"])
	#Divisão em treino e teste
	X_train, X_test, Y_train, Y_test = train_test_split(df["Perguntas"],df["Class"], test_size = 0.3, random_state = 42)
	#print(X_train.shape)
	#print(Y_train.shape)
	#print(X_test.shape)
	#print(Y_test.shape)



	vect = TfidfVectorizer().fit(X_train)

	with open("Models/vect_tresclasses", 'wb') as fid:
		pickle.dump(vect, fid)	

	X_train_vectorized = vect.transform(X_train)

	X_train_vectorized, Y_train = escala(X_train_vectorized, Y_train)

	clf = svm.SVC(gamma="scale",kernel='linear', degree=16,probability=True)
	clf.fit(X_train_vectorized,Y_train)

	with open(model_name, 'wb') as fid:
		pickle.dump(clf, fid)	

	preds = clf.predict(vect.transform(X_test))
	score = classification_report(Y_test, preds,target_names = ["Espaço Empresa","Apoios Sociais","RJACSR","Alojamento Local"])
	print(score)

	return vect

def corre_para_testes_restantes(modelo,frases,tresh_hold):
	with open(modelo, 'rb') as fid:
		clfrNB = pickle.load(fid)

	with open("Models/vect_tresclasses", 'rb') as fid:
		v = pickle.load(fid)

	sentences = []
	true_results = []
	for line in frases:
		line = line.replace("\n","")
		line = line.split("§")
		sentences.append(line[0])
		true_results.append(int(line[1]))
	print("Entrada recebida.")
	transformada = v.transform(sentences)
	preds = clfrNB.predict(transformada)
	preds_probs = clfrNB.predict_proba(transformada)

	#score = classification_report(true_results, preds,labels=[1,10,11])
	#print(score)
	'''
	all_classes = []
	all_valores = []
	max_prob_correto = []
	max_prob_errado = []
	for index, elem in enumerate(sentences):
		current_prob = preds_probs[index]
		if(preds[index]==true_results[index]):
			max_prob_correto.append(current_prob[current_prob.argmax()])
		else:
			max_prob_errado.append(current_prob[current_prob.argmax()])
		classes = []
		valores = []
		for index_2,elem in enumerate(current_prob):
			if(elem>tresh_hold):
				classes.append(index_2)
				valores.append(elem)
		frase = sentences[index] + "Prediction: "
		for elem in classes:
			frase+=" "+str(elem)+" "
		frase+=" probability " 
		for elem in valores:
			frase+=" "+str(elem)
		all_classes.append(classes)
		all_valores.append(valores)
		#print(frase)
	'''
	score = classification_report(true_results, preds)
	#target_names = ["Espaço Empresa","Apoios Sociais","RJACSR","Alojamento Local"]
	print(score)
	#hist = np.histogram([max_prob], bins=100)
	#print(max_prob)
	#plt.hist([max_prob_correto,max_prob_errado], bins=100)
	#plt.show()
	return preds

def corre_para_frase_restantes(modelo,frases,tresh_hold):
	with open(modelo, 'rb') as fid:
		clfrNB = pickle.load(fid)

	with open("Models/vect", 'rb') as fid:
		v = pickle.load(fid)

	sentences = []
	true_results = []
	sentences.append(frases[0].replace("\n",""))
	print("Entrada recebida.")
	transformada = v.transform(sentences)
	preds = clfrNB.predict(transformada)
	preds_probs = clfrNB.predict_proba(transformada)

	max_prob_correto = []
	max_prob_errado = []
	for index, elem in enumerate(sentences):
		current_prob = preds_probs[index]
		classes = []
		valores = []
		for index_2,elem in enumerate(current_prob):
			if(elem>tresh_hold):
				classes.append(index_2)
				valores.append(elem)
	return classes,valores

def corre(modelo,v):
	#RJACSR,AL,PE
	with open(modelo, 'rb') as fid:
		clfrNB = pickle.load(fid)

	with open("Models/vect_tresclasses", 'rb') as fid:
		v = pickle.load(fid)
	a = 0

	while (a==0):
		print("Entrada.")
		entrada = input()
		entrada = [entrada]
		#entrada = pd.DataFrame([entrada])
		print("Entrada recebida.")
		transformada = v.transform(entrada)
		preds = clfrNB.predict(transformada)
		preds_probs = clfrNB.predict_proba(transformada)
		print(preds)
		if(preds[0]==1):
			print("PE")
		elif(preds[0]==10):
			print("RJACSR")
		elif(preds[0]==11):
			print("AL")
		else:
			print("OOD")

def corre_modelo_real(modelo,frases,tresh_hold):
	
	with open(modelo, 'rb') as fid:
		clfrNB = pickle.load(fid)

	with open("Models/vect_tresclasses", 'rb') as fid:
		v = pickle.load(fid)

	sentences = frases
	transformada = v.transform(sentences)
	preds = clfrNB.predict(transformada)
	preds_probs = clfrNB.predict_proba(transformada)
	max_prob_correto = []
	max_prob_errado = []
	for index, elem in enumerate(sentences):
		current_prob = preds_probs[index]
		classes = []
		valores = []
		for index_2,elem in enumerate(current_prob):
			#print(elem)
			if(elem>tresh_hold):
				classes.append(index_2)
				valores.append(elem)
	return classes,valores
	'''
	for index,elem in enumerate(preds_probs):
		#print(sentences[index] + " "+ str(elem))
		print(elem)
	return preds_probs
	'''

def corre_modelo_completo(modelo_binario,modelo_conjunto,frase,tresh_hold):
	frases_validas = corre_para_frase(modelo_binario,frase)
	return corre_modelo_real(modelo_conjunto,frases_validas,tresh_hold)

if __name__ == '__main__':
	##########
	#Descomentar para treinar modelo multiclasses (1-4)
	##########
	#vect = treina("Models/svm_muticlass.pickle")

	#########
	#Coisas para testes
	#########
	'''
	frases = []
	print("##################VERSAO VG1#################")
	with open("Input/TestMulticlass/VG1.txt","r") as f:
		for line in f:
			frases.append(line)
			#print(line)
	corre_para_testes_restantes("Models/svm_muticlass.pickle",frases,0.8)

	frases = []
	print("##################VERSAO VG2#################")
	with open("Input/TestMulticlass/VG2.txt","r") as f:
		for line in f:
			frases.append(line)
			#print(line)
	corre_para_testes_restantes("Models/svm_muticlass.pickle",frases,0.8)

	frases = []
	print("##################VERSAO VUC#################")
	with open("Input/TestMulticlass/VUC.txt","r") as f:
		for line in f:
			frases.append(line)
			#print(line)
	corre_para_testes_restantes("Models/svm_muticlass.pickle",frases,0.8)

	frases = []
	print("##################VERSAO VIN#################")
	with open("Input/TestMulticlass/VIN.txt","r") as f:
		for line in f:
			frases.append(line)
			#print(line)
	corre_para_testes_restantes("Models/svm_muticlass.pickle",frases,0.8)
	'''
	'''
	frases = []
	print("##################VERSAO VIN#################")
	with open("Input/TestMulticlass/total.txt","r") as f:
		for line in f:
			frases.append(line)
			#print(line)
	corre_para_testes_restantes("Models/svm_muticlass.pickle",frases,0.8)
	'''

	#########
	#Correr binário seguido de normal
	#########
	tresh_hold = 0.9
	for i in range(10000):
		for elem in (corre_modelo_completo("Models/svm_binaria_v3.pickle","Models/svm_muticlass.pickle","frase de teste",tresh_hold)):
			print(elem)



	


import os
import io
import requests
import shutil
import zipfile

from ASAPPpy import ROOT_PATH

def get_confirm_token(response):
    '''
    Parameters
    ----------


    '''

    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None

def download(gid='1kR0hfuuaduTFixY0ZLgqkLRCw_ET5xRW', save_path=ROOT_PATH):
    '''
    Parameters
    ----------


    '''

    url = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(url, params={'id':gid}, stream=True)
    token = get_confirm_token(response)

    if token:
        params = {'id':gid, 'confirm':token}
        response = session.get(url, params=params, stream=True, headers={'Accept-Encoding': None})

    if 'Content-Length' in response.headers:
        length = response.headers.get('Content-Length')
    else:
        # read directly from the raw urllib3 connection
        print("Retrieving file size...")
        raw_content = response.raw.read()
        length = len(raw_content)
        # replace the internal file-object to serve the data again
        response.raw._fp = io.BytesIO(raw_content)

    if length:
        length = int(length)
        blocksize = max(4096, length//100)
        buffer = io.BytesIO()
        current_progress = 0

        print("Downloading data...")
        for data in response.iter_content(blocksize):
            current_progress += len(data)
            buffer.write(data)
            done = int(50 * current_progress / length)
            print("[%s%s]" % ('=' * done, ' ' * (50-done)), end='\r')

        print("\nUnzipping file...")
        zip_file = zipfile.ZipFile(buffer)

        if save_path is None:
            zip_file.extractall()
        else:
            zip_file.extractall(save_path)
    else:
        print("Could not find the length of the file. Downloading without showing progress bar and unzipping.")

        zip_file = zipfile.ZipFile(io.BytesIO(response.content))

        if save_path is None:
            zip_file.extractall()
        else:
            zip_file.extractall(save_path)

    # check if __MACOSX was created after unzipping and delete it
    if save_path is None:
        macosx_folder_path = '__MACOSX'
    else:
        macosx_folder_path = os.path.join(save_path, '__MACOSX')
    if os.path.exists(macosx_folder_path):
        shutil.rmtree(macosx_folder_path)

if __name__ == "__main__":
    # TAKE ID FROM SHAREABLE LINK
    file_id = '1kR0hfuuaduTFixY0ZLgqkLRCw_ET5xRW'
    # DESTINATION FILE ON YOUR DISK
    destination = None
    download(file_id, destination)


import os
import copy
import re
import asyncio
import slack
import pandas as pd

from gensim.parsing.preprocessing import strip_non_alphanum
from gensim.parsing.preprocessing import strip_multiple_whitespaces
from sklearn.metrics.pairwise import cosine_similarity
from nltk.corpus import stopwords
from string import punctuation

from ASAPPpy.feature_extraction import load_embeddings_models
import ASAPPpy.indexers.Whoosh.whoosh_make_query as qwi
from ASAPPpy.sts_model import STSModel

# constants
SLACK_SIGNING_SECRET = os.environ['SLACK_SIGNING_SECRET']
SLACK_BOT_TOKEN = os.environ['COBAIA_BOT_TOKEN']

# cobaiabot's user ID in Slack: value is assigned after the bot starts up
bot_id = None

# constants
MENTION_REGEX = "^<@(|[WU].+?)>(.*)"

word2vec_model, fasttext_model, ptlkb64_model, glove300_model, numberbatch_model = load_embeddings_models()

def pre_selection(unselected_phrases, model, position_correct_match):
	""" Function used to perform pre-selection """
	embeddings_unselected_phrases = []
	similaritires = []
	selected_phrases = []
	l_unselected_phrases = copy.deepcopy(unselected_phrases)

	for pair in l_unselected_phrases:
		temp_question = [model[word] for word in pair[0] if word in model.vocab]
		temp_variant = [model[word] for word in pair[1] if word in model.vocab]
		mean_question = sum(temp_question)/len(temp_question)
		mean_variant = sum(temp_variant)/len(temp_variant)

		embeddings_unselected_phrases.append([mean_question, mean_variant])

	for i in range(0, len(embeddings_unselected_phrases)):
		similarity = cosine_similarity([embeddings_unselected_phrases[i][0]], [embeddings_unselected_phrases[i][1]])
		similaritires.append(similarity[0][0])

	# compute the index of the 30 questions with the higher similarity
	selected_indexes = sorted(range(len(similaritires)), key=lambda x: similaritires[x])[-30:]

	# the correct match is added at the begining of the list
	if position_correct_match not in selected_indexes:
		selected_phrases = None
		selected_indexes = None
	else:
		selected_phrases.extend(l_unselected_phrases[position_correct_match])

		for j in range(len(selected_indexes)):
			if selected_indexes[j] != position_correct_match:
				selected_phrases.extend(l_unselected_phrases[selected_indexes[j]])
			else:
				selected_indexes.insert(0, selected_indexes.pop(j))

	return selected_phrases, selected_indexes

def n_max_elements(list1, N):
	""" Function to compute the N highest numbers of a list """
	n_list1 = copy.deepcopy(list1)
	final_list = []

	for i in range(0, N):
		max1 = 0

		for j in range(len(n_list1)):
			if n_list1[j] > max1:
				max1 = n_list1[j]

		n_list1.remove(max1)
		final_list.append(max1)

	return final_list

def chatbot_interface(interaction, word2vec_model, fasttext_model, ptlkb64_model, glove300_model, numberbatch_model):
	""" Function used to run the chatbot interface """

	# choose if stopwords should be removed from the user interaction
	process_interaction_toggle = 0

	# choose if pre-selection should not be used (0), used with word embeddings (1) or used with Whoosh (2)
	pre_selection_toggle = 2

	# parameters used to tune the selection of more than one response
	sr_alpha = 0.1
	sr_beta = 3

	# TODO: The STS model class can't perform feature selection but can use a model that uses it already.

	# location of the STS model
	model = STSModel()
	model.load_model('model_R_pos_adv-dependency_parsing-word2vec-ptlkb-numberbatch')

	faqs_variants_load_path = os.path.join('datasets', 'AIA-BDE_v2.0.txt')

	with open(faqs_variants_load_path) as faqs_file:
		faqs_variants_corpus = faqs_file.read().splitlines()

	faqs_file.close()

	faqs_variants_corpus = [line.replace('\t', '') for line in faqs_variants_corpus]
	faqs_variants_corpus = [line.split(':', 1) for line in faqs_variants_corpus]

	# add the original question to a different list to improve the conversational presentation of a response
	position = 0
	faqs_variants_questions = []

	for element in faqs_variants_corpus:
		if element[0] == 'P':
			faqs_variants_questions.append(element[1])
			position += 1 

	faqs_variants_corpus = [line for line in faqs_variants_corpus if len(line) == 2 and line[1]]
	faqs_variants_corpus = [[line[0], strip_non_alphanum(line[1])] if line[0] != 'R' else [line[0], line[1]] for line in faqs_variants_corpus]
	faqs_variants_corpus = [[line[0].rstrip(), line[1].rstrip()] if line[0] != 'R' else [line[0], line[1]] for line in faqs_variants_corpus]
	faqs_variants_corpus = [[line[0], strip_multiple_whitespaces(line[1])] if line[0] != 'R' else [line[0], line[1]] for line in faqs_variants_corpus]
	faqs_variants_corpus = [[line[0], line[1].lower()] if line[0] != 'R' else [line[0], line[1]] for line in faqs_variants_corpus]

	position = 0
	corpus = []

	for element in faqs_variants_corpus:
		if element[0] == 'P':
			corpus.append([element[1]])

		if element[0] == 'R':
			corpus[position].extend([element[1]])

			position += 1

	aux_list_of_questions = [phrases[0] for phrases in corpus]
	aux_df = pd.DataFrame(aux_list_of_questions, columns=['text'])

	# remove duplicate sentences from the aux_list_of_questions in order for Whoosh to work.
	clean_aux_list_of_questions = []

	for phrase in aux_list_of_questions:
		if phrase not in clean_aux_list_of_questions:
			clean_aux_list_of_questions.append(phrase)

	if process_interaction_toggle:
		print("The original sentence was: {}".format(interaction))
		stp = set(stopwords.words('portuguese') + list(punctuation))
		interaction = ' '.join([word for word in interaction.split(' ') if word not in stp])
		print("The sentenced after removing stopwords and punctuation: {}".format(interaction))

	unprocessed_corpus = []

	if 'response' not in aux_df:
		aux_df.insert(1, 'response', interaction)
	else:
		aux_df['response'] = interaction

	if pre_selection_toggle != 2:
		for j in range(len(aux_list_of_questions)):
			if pre_selection_toggle == 1:
				unprocessed_corpus.append([aux_list_of_questions[j], interaction])
			else:
				unprocessed_corpus.extend([aux_list_of_questions[j], interaction])

	if pre_selection_toggle == 1:
		corpus_pairs, indexes = pre_selection(unprocessed_corpus, fasttext_model, position)

		if corpus_pairs is None:
			index_path = os.path.join('indexers', 'Whoosh', 'indexes', 'cobaia_chitchat_v1.5')

			query_response = qwi.query_indexer(interaction, index_path)

			if (query_response[0] is None) or (not query_response[0]):
				response = "Desculpe, não percebi, pode colocar a sua questão de outra forma?"
				return response
			else:
				return response

		selected_aux_df = aux_df.iloc[indexes]
		selected_aux_df = selected_aux_df.reset_index(drop=True)
	else:
		if pre_selection_toggle == 2:
			pre_selection_index_path = os.path.join('indexers', 'Whoosh', 'indexes', 'FAQs_no_analyser_AIA-BDE_v2.0')

			query_response = qwi.query_indexer(interaction, pre_selection_index_path)
			options_docnumbers = query_response[2]

			if len(options_docnumbers) == 0:
				print("No options returned by the index.")

				for j in range(len(aux_list_of_questions)):
					unprocessed_corpus.extend([aux_list_of_questions[j], interaction])
			else:
				for pos, elem in enumerate(options_docnumbers):
					unprocessed_corpus.extend([clean_aux_list_of_questions[elem], interaction])

		corpus_pairs = unprocessed_corpus
		selected_aux_df = aux_df

	element_features = model.extract_multiple_features(corpus_pairs, 0, word2vec_mdl=word2vec_model, fasttext_mdl=fasttext_model, ptlkb_mdl=ptlkb64_model, glove_mdl=glove300_model, numberbatch_mdl=numberbatch_model)

	predicted_similarity = model.predict_similarity(element_features)
	predicted_similarity = predicted_similarity.tolist()

	highest_match = max(predicted_similarity)

	if highest_match < 2.5:
		# the query search will return a list of phrases with the highest matches, which will be used with the similarity model in order to evaluate which answer should be returned to the user
		index_path = os.path.join('indexers', 'Whoosh', 'indexes', 'cobaia_chitchat_v1.5')

		query_response = qwi.query_indexer(interaction, index_path)

		if (query_response[0] is None) or (not query_response[0]):
			response = "Desculpe, não percebi, pode colocar a sua questão de outra forma?"
			return response
		else:
			unprocessed_answers = []
			aux_qwi = pd.DataFrame(query_response[0], columns=['text'])

			if 'response' not in aux_qwi:
				aux_qwi.insert(1, 'response', interaction)
			else:
				aux_qwi['response'] = interaction

			for k in range(len(query_response[0])):
				unprocessed_answers.extend([aux_list_of_questions[k], interaction])

			element_features_qwi = model.extract_multiple_features(unprocessed_answers, 0, word2vec_mdl=word2vec_model, fasttext_mdl=fasttext_model, ptlkb_mdl=ptlkb64_model, glove_mdl=glove300_model, numberbatch_mdl=numberbatch_model)

			predicted_similarity_qwi = model.predict_similarity(element_features_qwi)
			predicted_similarity_qwi = predicted_similarity_qwi.tolist()

			highest_match_index_qwi = predicted_similarity_qwi.index(max(predicted_similarity_qwi))

			return query_response[1][highest_match_index_qwi]
	else:
		selectable_range = (max(predicted_similarity)-min(predicted_similarity)) * sr_alpha

		possible_matches = n_max_elements(predicted_similarity, sr_beta)

		highest_match_index = predicted_similarity.index(max(predicted_similarity))
		#should be index 1, for testing purposes it is 0
		response = ("Se a sua pergunta foi: %s \nR: %s\n" % (faqs_variants_questions[highest_match_index], corpus[highest_match_index][1]))

		for i in range(1, sr_beta):
			if abs(highest_match-possible_matches[i]) <= selectable_range:
				response += ("Também poderá estar interessado em: %s\nR: %s\n" % (faqs_variants_questions[predicted_similarity.index(possible_matches[i])], corpus[predicted_similarity.index(possible_matches[i])][1]))

		return response

@slack.RTMClient.run_on(event='message')
async def handle_message(**payload):

	data = payload['data']

	# get the text of the message in order to search for a response
	text = data.get('text')
	# get the channel in which the message was sent
	channel = data.get('channel')

	# if the first letter in the channel name is a D, it means that is was a direct message to the chatbot
	if channel[0] == "D":
		if data.get('subtype') is None and text:
			message = chatbot_interface(text, word2vec_model, fasttext_model, ptlkb64_model, glove300_model, numberbatch_model)

			webclient = payload['web_client']
			await webclient.chat_postMessage(
				channel=channel,
				text=message
			)
	else:
		# if not a direct message, check whether or not the chatbot was mentioned in the conversation
		user_id, text = parse_direct_mention(text)

		if user_id == bot_id:
			if data.get('subtype') is None and text:
				message = chatbot_interface(text, word2vec_model, fasttext_model, ptlkb64_model, glove300_model, numberbatch_model)

				webclient = payload['web_client']
				await webclient.chat_postMessage(
					channel=channel,
					text=message
				)

def parse_direct_mention(message_text):
	"""
		Finds a direct mention (a mention that is at the beginning) in message text
		and returns the user ID which was mentioned. If there is no direct mention, returns None
	"""
	matches = re.search(MENTION_REGEX, message_text)
	# the first group contains the username, the second group contains the remaining message
	return (matches.group(1), matches.group(2).strip()) if matches else (None, None)

if __name__ == "__main__":
	# TODO (Future Work): Verify if it is possible to use @mentions in a more elegant way
	slack_web_client = slack.WebClient(SLACK_BOT_TOKEN)

	# extract the bot_id in order to check if it is mentioned in a channel conversation
	if bot_id is None:
		bot_id = slack_web_client.api_call("auth.test")["user_id"]

	loop = asyncio.new_event_loop()
	asyncio.set_event_loop(loop)
	slack_client = slack.RTMClient(
		token=SLACK_BOT_TOKEN, run_async=True, loop=loop
	)
	loop.run_until_complete(slack_client.start())



from ASAPPpy.sentences.classes import Sentence

s = Sentence("O João gosta de jogar futebol e de vez em quando também joga basquete em Viseu quando o tempo está bom")

print(s.text)
print(s.lemmas)

# s.compute_all()
s.compute_tags()
s.compute_lemmas()
s.compute_entities()

print(s.text)
print(s.lemmas)
print(s.tags)
print(s.entities)



import os
import sys
import time
from whoosh.index import create_in
from whoosh.fields import Schema, TEXT, ID
from whoosh.analysis import StemmingAnalyzer, CharsetFilter, NgramFilter
from whoosh.support.charset import accent_map
from ASAPPpy import ROOT_PATH

def createSearchableData(directory, load_path):
    '''
    Schema definition: title(name of file), path(as ID), content(indexed
    but not stored),textdata (stored text content)
    '''
    # the call of the StemmingAnalyzer had to be changed in the whoosh directory to support the portuguese language
    my_analyzer = StemmingAnalyzer() | CharsetFilter(accent_map) | NgramFilter(minsize=2, maxsize=4)
    schema = Schema(question=TEXT(analyzer=my_analyzer, stored=True), response=TEXT(analyzer=my_analyzer, stored=True))
    # schema = Schema(question=TEXT(stored=True), response=TEXT(stored=True))
    schema.cachesize = -1

    if not os.path.exists(directory):

        # makedirs is used to create directories with subdirectories in it
        os.makedirs(directory)

    # Creating a index writer to add document as per schema
    ix = create_in(directory, schema)
    writer = ix.writer(limitmb=1024)

    with open(load_path) as subtles_file:
        subtles_corpus = subtles_file.read().splitlines()

    for i in range(0, len(subtles_corpus), 2):
        writer.add_document(question=subtles_corpus[i], response=subtles_corpus[i+1])
    writer.commit()

if __name__ == '__main__':

    directory = os.path.join('indexes', 'Subtle_removed_entities')
    load_path = os.path.join(ROOT_PATH, 'datasets', 'SubtleCorpusPTEN', 'por', 'corpus0sDialogues_clean_removed_entities.txt')

    start_time = time.time()
    print("Started creating the index")

    createSearchableData(directory, load_path)

    print("Finished creating index successfully")
    print("--- %s seconds ---" %(time.time() - start_time))
    print('\a')



'''
The Chatbot Module
'''

import os
import time

import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from random import randint
from joblib import load
from sklearn import metrics

from resources.resources import read_faqs_variants, read_class_set, pre_selection, n_max_elements, n_max_elements_indexes, select_multiple_random_subtitles, write_output_file
# from feature_extraction import extract_features
from tmp_feature_extraction import extract_features, load_features
from classifiers.svm import svm_classifier
from classifiers.svm_restantes_classes import corre_para_testes_restantes
from indexers.Whoosh.whoosh_make_query import query_indexer
from ASAPPpy.sts_model import STSModel

def chatbot(word2vec_model=None, fasttext_model=None, ptlkb64_model=None, ptlkb128_model=None, glove100_model=None, glove300_model=None, numberbatch_model=None):
	"""
	Function used to run the chatbot

	"""

	# Flag to indicate if classification should be used (1) or not (0)
	classification_flag = 0

	# Choose if pre-selection should not be used (0), used with word embeddings (1) or used with Whoosh (2)
	pre_selection_flag = 2

	# Choose whether (1) or not (0) a new file with VUC random positions should be generated
	generate_random_positions_file_flag = 0

	# load the pre-trained STS model
	model_name = 'model_1906_ablation_study_master'

	model = STSModel()
	model.load_model(model_name)

	if classification_flag:
		print("The classifier is being used.")

		# read the different class sets
		class_1, class_2, class_3 = read_class_set()

		# transform the class sets from lists to dataframes
		class_1_df = pd.DataFrame(class_1, columns=['text'])
		class_2_df = pd.DataFrame(class_2, columns=['text'])
		class_3_df = pd.DataFrame(class_3, columns=['text'])

	OG, VIN, VG1, VG2, VUC, VMT = read_faqs_variants()

	# to test the system different sets of FAQs' variants (VIN, VG1, VG2, VUC, VMT) can be used, as well as the original questions (OG). To do so, they should be in the form: all_variants = [OG, VIN, VG1, VG2, VUC, VMT] and all_variants_names = ["OG", "VIN", "VG1", "VG2", "VUC", "VMT"]
	all_variants = [OG, VIN, VG1, VG2, VUC, VMT]
	all_variants_names = ["OG", "VIN", "VG1", "VG2", "VUC", "VMT"]

	all_accuracies = [['Variant collection', 'Mean', 'Number of Correct matches', 'Total number of questions', "Top 3", "Top 5", "Threshold 0", "Threshold 1", "Threshold 1,5", "Threshold 2", "Threshold 2,5", "Threshold 3", "Threshold 3,5", "Threshold 4",]]

	if generate_random_positions_file_flag == 1:
		random_positions_VUC = []

	incorrect_matches_VG1 = []
	use_only_one_vuc = 0

	if use_only_one_vuc:
		with open('random_positions_VUC.txt') as random_positions_VUC_file:
			random_positions_VUC = random_positions_VUC_file.read().splitlines()

		random_positions_VUC_file.close()

	# given that que questions indexed with Whoosh were preprocessed, the original questions are needed in order to confirm if the correct response was returned
	if pre_selection_flag == 2:
		original_questions_with_duplicates = [phrases[0] for phrases in OG]

		original_questions = []
		for question in original_questions_with_duplicates:
			if question not in original_questions:
				original_questions.append(question)

	# used to plot the ROC curve
	y_true = []
	y_score = []

	for pos_variant, variant in enumerate(all_variants):

		if all_variants_names[pos_variant] == 'VIN':
			start_time = time.time()
			print("Started testing the chatbot")

		list_of_questions = [phrases[0] for phrases in variant]
		questions_df = pd.DataFrame(list_of_questions, columns=['text'])

		n_correct_matches = 0
		n_variant_questions = 0
		n_top3 = 0
		n_top5 = 0
		threshold_0 = 0
		threshold_1 = 0
		threshold_1_5 = 0
		threshold_2 = 0
		threshold_2_5 = 0
		threshold_3 = 0
		threshold_3_5 = 0
		threshold_4 = 0

		for position, element in enumerate(variant):
			print('Variant %s - %d/%d' % (all_variants_names[pos_variant], position+1, len(all_variants[pos_variant])), end='\r')

			for i in range(1, len(element)-1):
				unprocessed_corpus = []
				n_variant_questions += 1

				# select a random variant from VUC
				if all_variants_names[pos_variant] == 'VUC' and use_only_one_vuc:
					i = int(random_positions_VUC[position])

					if generate_random_positions_file_flag == 1:
						i = randint(1, len(element)-2)
						random_positions_VUC.append(i)
						i = 1

				if classification_flag:
					# apply the classifier before using the STS model
					# predicted_class = svm_classifier(element[i])
					predicted_class = corre_para_testes_restantes([element[i]])

					if predicted_class == 1:
						aux_df = class_1_df
						aux_list_of_questions = class_1
					elif predicted_class == 2:
						aux_df = class_2_df
						aux_list_of_questions = class_2
					elif predicted_class == 3:
						aux_df = class_3_df
						aux_list_of_questions = class_3
					else:
						aux_df = questions_df
						aux_list_of_questions = list_of_questions
				else:
					aux_df = questions_df
					aux_list_of_questions = list_of_questions

				if 'response' not in aux_df:
					aux_df.insert(1, 'response', element[i])
				else:
					aux_df['response'] = element[i]

				if pre_selection_flag != 2:
					for j in range(len(aux_list_of_questions)):
						if int(pre_selection_flag) == 1:
							unprocessed_corpus.append([aux_list_of_questions[j], element[i]])
						else:
							unprocessed_corpus.extend([aux_list_of_questions[j], element[i]])

				if pre_selection_flag == 1:
					corpus_pairs, indexes = pre_selection(unprocessed_corpus, fasttext_model, position)

					if corpus_pairs is None:
						if all_variants_names[pos_variant] == 'VUC' and use_only_one_vuc:
							break
						continue

					selected_aux_df = aux_df.iloc[indexes]
					selected_aux_df = selected_aux_df.reset_index(drop=True)
				else:
					if pre_selection_flag == 2:
						index_path = os.path.join('indexers', 'Whoosh', 'indexes', 'FAQs_stemming_analyser_charset_filter_AIA-BDE_v2.0_no_preprocessing')

						query_response = query_indexer(element[i], index_path)
						options_docnumbers = query_response[2]

						if len(options_docnumbers) == 0:
							continue
						
						aux_corpus_pairs = []

						for pos, elem in enumerate(options_docnumbers):
							unprocessed_corpus.extend([original_questions[elem], element[i]])
							#aux_corpus_pairs.append([original_questions[elem], element[i]])
							aux_corpus_pairs.append(original_questions[elem])

					corpus_pairs = unprocessed_corpus
					selected_aux_df = aux_df

				element_features = model.extract_multiple_features(corpus_pairs, 0, word2vec_mdl=word2vec_model, fasttext_mdl=fasttext_model, ptlkb_mdl=ptlkb64_model, glove_mdl=glove300_model, numberbatch_mdl=numberbatch_model)

				# number_of_features_train = converted_mask.count(1)
				# number_of_features_test = len(element_features[0])

				# if number_of_features_train > number_of_features_test:
				# 	element_features_list = element_features.tolist()
				# 	for feature_pair in element_features_list:
				# 		feature_pair.extend([0]*(number_of_features_train-number_of_features_test))
				# 	element_features = np.asarray(element_features_list)

				predicted_similarity = model.predict_similarity(element_features)
				predicted_similarity = predicted_similarity.tolist()

				highest_match = max(predicted_similarity)
				highest_match_index = predicted_similarity.index(max(predicted_similarity))

				# if pre_selection is used the correct match will always be in the first position
				if int(pre_selection_flag) == 1:
					if highest_match_index == 0:
						n_correct_matches += 1
						if 0 <= highest_match < 1:
							threshold_0 += 1
						if 1 <= highest_match < 1.5:
							threshold_1 += 1
						if 1.5 <= highest_match < 2:
							threshold_1_5 += 1
						if 2 <= highest_match < 2.5:
							threshold_2 += 1
						if 2.5 <= highest_match < 3:
							threshold_2_5 += 1
						if 3 <= highest_match < 3.5:
							threshold_3 += 1
						if 3.5 <= highest_match < 4:
							threshold_3_5 += 1
						if 4 <= highest_match < 5:
							threshold_4 += 1
					else:
						top3 = n_max_elements(predicted_similarity, 3)
						top5 = n_max_elements(predicted_similarity, 5)

						if predicted_similarity[0] in top3:
							n_top3 += 1

						if (predicted_similarity[0] in top5) and (predicted_similarity[0] not in top3):
							n_top5 += 1

						# for research purposes only. Used to evaluate why mismatches happen with VG1
						if all_variants_names[pos_variant] == 'VG1':
							incorrect_matches_VG1.append([element[0], element[i], aux_list_of_questions[highest_match_index], highest_match])
				else:
					# if highest_match_index == position:
					if pre_selection_flag == 0:
						match = aux_list_of_questions[highest_match_index]
					elif pre_selection_flag == 2:
						match = aux_corpus_pairs[highest_match_index]
					else:
						print("Pre-selection flag was assigned a wrong number. Please fix it and re-run the script.")

					if match == element[0]:
						# used to plot the ROC curve
						y_true.append(1)
						y_score.append(highest_match)

						n_correct_matches += 1
						if 0 <= highest_match < 1:
							threshold_0 += 1
						if 1 <= highest_match < 1.5:
							threshold_1 += 1
						if 1.5 <= highest_match < 2:
							threshold_1_5 += 1
						if 2 <= highest_match < 2.5:
							threshold_2 += 1
						if 2.5 <= highest_match < 3:
							threshold_2_5 += 1
						if 3 <= highest_match < 3.5:
							threshold_3 += 1
						if 3.5 <= highest_match < 4:
							threshold_3_5 += 1
						if 4 <= highest_match < 5:
							threshold_4 += 1
					else:
						# used to plot the ROC curve
						y_true.append(0)
						y_score.append(highest_match)

						if pre_selection_flag == 0:
							top3 = n_max_elements(predicted_similarity, 3)
							top5 = n_max_elements(predicted_similarity, 5)

							if predicted_similarity[position] in top3:
								n_top3 += 1

							if (predicted_similarity[position] in top5) and (predicted_similarity[position] not in top3):
								n_top5 += 1

							# for research purposes only. Used to evaluate why mismatches happen with VG1
							if all_variants_names[pos_variant] == 'VG1':
								incorrect_matches_VG1.append([element[0], element[i], aux_list_of_questions[highest_match_index], highest_match])
						
						elif pre_selection_flag == 2:
							if len(predicted_similarity) >= 3:
								top3 = n_max_elements_indexes(predicted_similarity, 3)
							else:
								continue

							if len(predicted_similarity) >= 5:
								top5 = n_max_elements_indexes(predicted_similarity, 5)
							else:
								continue

							for index in top3:
								if aux_corpus_pairs[index] == element[0]:
									n_top3 += 1

							for index in top5:
								if aux_corpus_pairs[index] == element[0] and index not in top3:
									n_top5 += 1
						
						else:
							print("Pre-selection flag was assigned a wrong number. Please fix it and re-run the script.")


						'''
						top3 = n_max_elements_indexes(predicted_similarity, 3)
						top5 = n_max_elements_indexes(predicted_similarity, 5)

						current_value_top3 = n_top3

						for index in top3:
							# print(aux_list_of_questions[index])
							if aux_list_of_questions[index] == element[0]:
								# print("In top 3")
								n_top3 += 1

						for index in top5:
							# print(aux_list_of_questions[index])
							if aux_list_of_questions[index] == element[0] and current_value_top3 != n_top3:
								# print("In top 5")
								n_top5 += 1
						'''
						

				# uncomment if only pretends to evaluate one variant question of VUC
				if all_variants_names[pos_variant] == 'VUC' and use_only_one_vuc:
					break

			if len(all_variants[pos_variant]) == position + 1:
				print()

		variant_accuracy = (n_correct_matches/n_variant_questions)*100

		all_accuracies.append([all_variants_names[pos_variant], variant_accuracy, n_correct_matches, n_variant_questions, n_top3, n_top5, threshold_0, threshold_1, threshold_1_5, threshold_2, threshold_2_5, threshold_3, threshold_3_5, threshold_4])

	'''
	# used to plot the ROC curve
	y = np.array(y_true)
	scores = np.array(y_score)
	fpr, tpr, thresholds = metrics.roc_curve(y, scores)

	print("FPR: ")
	print(fpr)
	print("TPR: ")
	print(tpr)
	print("Threshholds: ")
	print(thresholds)

	# finding the optimal threshold
	optimal_idx = np.argmax(tpr - fpr)
	print(optimal_idx)
	optimal_threshold = thresholds[optimal_idx]
	print(optimal_threshold)

	# finding the N optimal thresholds
	array_of_differences = tpr-fpr
	list_of_differences = array_of_differences.tolist()

	optimal_indexes = n_max_elements_indexes(list_of_differences, 10)

	list_of_optimal_thresholds = []

	for index in optimal_indexes:
		list_of_optimal_thresholds.append(thresholds[index])

	print("The list of optimal thresholds is: ")
	print(list_of_optimal_thresholds)

	plt.figure()
	lw = 2
	plt.plot(fpr, tpr, color='darkorange', lw=lw, label='ROC curve')
	plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
	plt.plot(fpr[optimal_idx], tpr[optimal_idx], color='red', marker='o', markersize=5, label='Optimal Threshold')
	plt.xlim([0.0, 1.0])
	plt.ylim([0.0, 1.0])
	plt.xlabel('False Positive Rate')
	plt.ylabel('True Positive Rate')
	plt.title('Similarity Thresholds ROC curve')
	plt.legend(loc="lower right")
	plt.show()
	plt.close()
	'''
	# writing VG1 mismatches to file
	with open('vg1_mismatches.txt', 'w') as mismatches_file:
		for item in incorrect_matches_VG1:
			mismatches_file.write("%s\n" % item)

	mismatches_file.close()

	total_correct_matches = 0
	total_variant_questions = 0
	total_top3 = 0
	total_top5 = 0

	for accuracy in all_accuracies:
		if accuracy[0] != "OG":
			if isinstance(accuracy[2], int):
				total_correct_matches += accuracy[2]
				total_variant_questions += accuracy[3]
				total_top3 += accuracy[4]
				total_top5 += accuracy[5]

	total_accuracy = (total_correct_matches/total_variant_questions)*100

	all_accuracies.extend([['Variant collection', 'Mean', 'Total Top3', 'Total Top5'], ["ALL", total_accuracy, total_top3, total_top5]])

	if generate_random_positions_file_flag == 1:
		with open('random_positions_VUC.txt', 'w') as random_positions_VUC_file:
		    for item in random_positions_VUC:
		        random_positions_VUC_file.write("%s\n" % item)

		random_positions_VUC_file.close()

	# writing the results to file
	write_output_file(model_name, all_accuracies, classification_flag, model.feature_selection, pre_selection_flag)

	print("Testing finished successfully")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')




'''
Module used to extract syntactic features from a given corpus.
'''

import spacy
import pandas as pd

from .lexical_features import jaccard_coefficient

def compute_pos(pipe_tags):
	""" Function used to compute the POS-tags difference between two sentences in the corpus """

	tags = ['adj', 'adv', 'art', 'conj-c', 'conj-s', 'intj', 'n', 'n-adj', 'num', 'pron-det', 'pron-indp', 'pron-pers', 'prop', 'prp', 'punc', 'v-fin', 'v-ger', 'v-inf', 'v-pcp']

	prepocessed_tagged_corpus = [line.split(' ') for line in pipe_tags]

	tagged_corpus_counts = []

	for i in range(0, len(prepocessed_tagged_corpus), 2):
		tagged_pair_counts = []

		# all_tags_difference = abs(len(prepocessed_tagged_corpus[i])-len(prepocessed_tagged_corpus[i+1]))
		# tagged_pair_counts.append(all_tags_difference)

		for tag in tags:
			temp_q = prepocessed_tagged_corpus[i].count(tag)
			temp_r = prepocessed_tagged_corpus[i+1].count(tag)

			difference = abs(temp_q-temp_r)
			tagged_pair_counts.append(difference)

		tagged_corpus_counts.append(tagged_pair_counts)

	# tags.insert(0, "all_tags")

	# convert list to Dataframe in order to use it with feature extraction
	tagged_corpus_counts = pd.DataFrame(tagged_corpus_counts)
	tagged_corpus_counts.columns = tags

	return tagged_corpus_counts

def dependency_parsing(parse_corpus):
	""" Function used to compute the syntactic dependencies between two sentences in the corpus """

	spacy_nlp = spacy.load('pt')

	dependency_values = []

	for i in range(0, len(parse_corpus), 2):
		sentence1_triplets = []
		sentence2_triplets = []

		parsed_sentence_1 = spacy_nlp(parse_corpus[i])
		parsed_sentence_2 = spacy_nlp(parse_corpus[i+1])

		for token in parsed_sentence_1:
			if token.dep_ != "ROOT" and token.dep_ != "punct":
				sentence1_triplets.append((token.text, token.head.text, token.dep_))

		for token in parsed_sentence_2:
			if token.dep_ != "ROOT" and token.dep_ != "punct":
				sentence2_triplets.append((token.text, token.head.text, token.dep_))

		pair_jc = jaccard_coefficient(set(sentence1_triplets), set(sentence2_triplets))

		dependency_values.append(pair_jc)

	# convert list to Dataframe in order to use it with feature extraction
	dependency_parsing_jc = pd.DataFrame(dependency_values)
	dependency_parsing_jc.columns = ["dependency_parsing_jc"]

	return dependency_parsing_jc
	


'''
Module used to compute similarity between sentences of the ASSIN collection using word embeddings.
'''

import time
import os
from xml.etree import cElementTree as ET
from gensim.models import KeyedVectors
from sklearn.metrics.pairwise import cosine_similarity

#import function to read the xml file
#from assin.assineval.commons import read_xml
from ...scripts.tools import preprocessing, compute_tfidf_matrix

global n_iterations
n_iterations = 0

def apply_tfidf_model(sentence, embedding_model, tfidf_model):
	""" Auxiliar function used to compute the vector of each sentence by applying the word2vec model and the tfidf model """

	global n_iterations

	updated_sentence = []

	for word in sentence:
		if word in embedding_model.vocab:
			new_word = embedding_model[word] * tfidf_model[n_iterations][word]
			updated_sentence.append(new_word)

	n_iterations += 2
	return updated_sentence

def compute_models(model, embeddings_data, tfidf_data, use_tf_idf=0, rm_stopwords=0, numbers_to_text=0):
	""" Function used to apply the word embeddings model and compute the vector of each sentence """

	global n_iterations

	if use_tf_idf == 1:

		# tokenization should be 0 when applying preprocessing to the data that will be used to build the tfidf matrix
		tfidf_preprocessed = preprocessing(tfidf_data, 0, rm_stopwords, numbers_to_text, use_tf_idf)
		tfidf_model = compute_tfidf_matrix(tfidf_preprocessed, rm_stopwords, 1, 0)

		n_iterations = 0
		embeddings_data['text'] = embeddings_data['text'].apply(lambda x: apply_tfidf_model(x, model, tfidf_model))
		n_iterations = 1
		embeddings_data['response'] = embeddings_data['response'].apply(lambda x: apply_tfidf_model(x, model, tfidf_model))
		n_iterations = 0
	else:
		# The NILC embeddings don't allow to increment the existing model with new data, because it
		# doesn't have a binary version of it.

		embeddings_data['text'] = embeddings_data['text'].apply(lambda x: [model[word] for word in x if word in model.vocab])
		embeddings_data['response'] = embeddings_data['response'].apply(lambda x: [model[word] for word in x if word in model.vocab])

	#compute the mean of the sentence
	# embeddings_data['text'] = embeddings_data['text'].apply(lambda x: sum(x)/len(x))
	# embeddings_data['response'] = embeddings_data['response'].apply(lambda x: sum(x)/len(x))

	#uncomment if using all corpus for test and training is needed
	embeddings_data['text'] = embeddings_data['text'].apply(lambda x: sum(x)/len(x) if len(x) != 0 else [0]*64)
	embeddings_data['response'] = embeddings_data['response'].apply(lambda x: sum(x)/len(x) if len(x) != 0 else [0]*64)

	return embeddings_data

def ptlkb_model(model, tf_idf, remove_stopwords, convert_num_to_text, pipe_lemmas):
	""" Function used to load the word embeddings model and compute the vector of each sentence """

	word_embeddings = []

	# if run_pipeline == 0:
	# 	if system_mode == 0:
	# 		lemmas_path = os.path.join('dataset', 'FAQ_todas_variantes_texto_lemmatized.txt')
	# 	elif system_mode == 1:
	# 		lemmas_path = os.path.join('NLPyPort', 'assin-ptpt-ptbr-train-lemmatized.txt')
	# 	elif system_mode == 2:
	# 		lemmas_path = os.path.join('NLPyPort', 'assin-ptpt-ptbr-train-test-lemmatized.txt')
	# 	elif system_mode == 3:
	# 		lemmas_path = os.path.join('dataset', 'SubtleCorpusPTEN', 'por', 'corpus0sDialogues_clean_lemmatized.txt')

	# 	with open(lemmas_path) as lemmatized_file:
	# 		lemmatized_corpus = lemmatized_file.read().splitlines()

	# 	lemmatized_file.close()
	# else:
	lemmatized_corpus = pipe_lemmas

	# when it comes to word embeddings, the preprocessing function should always have tokenization equal to 1
	preprocessed_corpus = preprocessing(lemmatized_corpus, 1, remove_stopwords, convert_num_to_text, tf_idf)

	if tf_idf == 1:
		embeddings_data = compute_models(model, preprocessed_corpus, lemmatized_corpus, tf_idf, remove_stopwords, convert_num_to_text)
	else:
		embeddings_data = compute_models(model, preprocessed_corpus, lemmatized_corpus, tf_idf)

	for i in range(len(embeddings_data)):
		similarity = cosine_similarity([embeddings_data['text'][i]], [embeddings_data['response'][i]])
		word_embeddings.append(similarity[0][0])

	return word_embeddings




'''
Module used for feature selection.
'''

import os

from sklearn.ensemble import RandomForestRegressor
from sklearn.feature_selection import SelectPercentile, SelectFromModel, RFE, f_regression
from ASAPPpy.scripts.tools import write_data_to_file

def build_mask(mask, used_features):
	""" Function used to complete the mask with unused features not available for feature selection """

	for index, key in enumerate(used_features):
		if not mask[index]:
			used_features[key] = 0

	return used_features

def best_percentile_selector(train_features, test_features, train_similarity_target, test_similarity_target, regressor):
	""" Function used to select the best percentile selector """
	percentile_score = 0
	percentiles = [25, 35, 45, 50, 55, 65, 75]
	# percentiles = [45]
	percentile_selector = None
	percentile_train_features_selected = None
	percentile_test_features_selected = None

	for percentile in percentiles:
		print(percentile)
		temp_percentile_selector = SelectPercentile(score_func=f_regression, percentile=percentile)
		temp_percentile_selector.fit(train_features, train_similarity_target)
		temp_percentile_train_features_selected = temp_percentile_selector.transform(train_features)
		temp_percentile_test_features_selected = temp_percentile_selector.transform(test_features)

		regressor.fit(temp_percentile_train_features_selected, train_similarity_target)

		temp_score = regressor.score(temp_percentile_test_features_selected, test_similarity_target)
		print("The score on the selected features (Percentile Selector): %.3f" % temp_score)

		if temp_score > percentile_score:
			percentile_score = temp_score
			percentile_selector = temp_percentile_selector
			percentile_train_features_selected = temp_percentile_train_features_selected
			percentile_test_features_selected = temp_percentile_test_features_selected

	percentile_mask = percentile_selector.get_support()
	print("This is the percentile mask: ")
	print(percentile_mask)

	return percentile_selector, percentile_score, percentile_train_features_selected, percentile_test_features_selected, percentile_mask

def best_model_based_selector(train_features, test_features, train_similarity_target, test_similarity_target, regressor):
	""" Function used to select the best model based selector """
	model_based_score = 0
	scaling_factors = ["0.25*mean", "0.5*mean", "median", "1.25*mean", "1.5*mean"]
	# scaling_factors = ["0.5*mean", "median"]
	model_based_selector = None
	model_based_train_features_selected = None
	model_based_test_features_selected = None

	for factor in scaling_factors:
		print(factor)
		temp_model_based_selector = SelectFromModel(RandomForestRegressor(n_estimators=100), threshold=factor)
		temp_model_based_selector.fit(train_features, train_similarity_target)
		temp_model_based_train_features_selected = temp_model_based_selector.transform(train_features)
		temp_model_based_test_features_selected = temp_model_based_selector.transform(test_features)

		regressor.fit(temp_model_based_train_features_selected, train_similarity_target)

		temp_score = regressor.score(temp_model_based_test_features_selected, test_similarity_target)
		print("The score on the selected features (Model Based Selector): %.3f" % temp_score)

		if temp_score > model_based_score:
			model_based_score = temp_score
			model_based_selector = temp_model_based_selector
			model_based_train_features_selected = temp_model_based_train_features_selected
			model_based_test_features_selected = temp_model_based_test_features_selected

	model_based_mask = model_based_selector.get_support()
	print("This is the model based mask: ")
	print(model_based_mask)

	return model_based_selector, model_based_score, model_based_train_features_selected, model_based_test_features_selected, model_based_mask

def best_iterative_based_selector(train_features, test_features, train_similarity_target, test_similarity_target, regressor):
	""" Function used to select the best iterative based selector """
	iterative_based_score = 0
	# given that all pairs use the same amount of features, the position 0 was arbitrarily selected to compute the number of features being used
	min_number_features = int(0.15*len(train_features[0]))
	max_number_features = int(0.85*len(train_features[0]))

	# min_number_features = 19
	# max_number_features = 20

	iterative_based_selector = None
	iterative_based_train_features_selected = None
	iterative_based_test_features_selected = None

	for i in range(min_number_features, max_number_features):
		print(i)
		temp_iterative_based_selector = RFE(RandomForestRegressor(n_estimators=100), n_features_to_select=i)
		temp_iterative_based_selector.fit(train_features, train_similarity_target)
		temp_iterative_based_train_features_selected = temp_iterative_based_selector.transform(train_features)
		temp_iterative_based_test_features_selected = temp_iterative_based_selector.transform(test_features)

		regressor.fit(temp_iterative_based_train_features_selected, train_similarity_target)

		temp_score = regressor.score(temp_iterative_based_test_features_selected, test_similarity_target)
		print("The score on the selected features (Iterative Based Selector): %.3f" % temp_score)

		if temp_score > iterative_based_score:
			iterative_based_score = temp_score
			iterative_based_selector = temp_iterative_based_selector
			iterative_based_train_features_selected = temp_iterative_based_train_features_selected
			iterative_based_test_features_selected = temp_iterative_based_test_features_selected

	iterative_based_mask = iterative_based_selector.get_support()
	print("This is the iterative based mask: ")
	print(iterative_based_mask)

	return iterative_based_selector, iterative_based_score, iterative_based_train_features_selected, iterative_based_test_features_selected, iterative_based_mask

def feature_selection(train_features, test_features, train_similarity_target, test_similarity_target, regressor, used_features):
	# TODO: Check why were test_features being returned
	""" Function used to perform feature selection """
	# percentile selector
	percentile_selector, percentile_score, percentile_train_features_selected, percentile_test_features_selected, percentile_mask = best_percentile_selector(train_features, test_features, train_similarity_target, test_similarity_target, regressor)

	# model based selector
	model_based_selector, model_based_score, model_based_train_features_selected, model_based_test_features_selected, model_based_mask = best_model_based_selector(train_features, test_features, train_similarity_target, test_similarity_target, regressor)

	# iterative based selector
	iterative_based_selector, iterative_based_score, iterative_based_train_features_selected, iterative_based_test_features_selected, iterative_based_mask = best_iterative_based_selector(train_features, test_features, train_similarity_target, test_similarity_target, regressor)

	all_scores = []

	regressor.fit(train_features, train_similarity_target)
	print("The score on all features: %.3f" % regressor.score(test_features, test_similarity_target))
	all_scores.append(regressor.score(test_features, test_similarity_target))

	# show results for the percentile selector
	all_scores.append(percentile_score)

	# show results for the model based selector
	all_scores.append(model_based_score)

	# show results for the iterative based selector
	all_scores.append(iterative_based_score)

	max_value_position = all_scores.index(max(all_scores))

	if max_value_position == 0:
		print("Returning all features!\n")
		return None
	elif max_value_position == 1:
		percentile_mask = build_mask(percentile_mask, used_features)
		print(percentile_mask)

		print("Returning features selected with the percentile selector!\n")
		return percentile_selector, percentile_train_features_selected, percentile_mask
	elif max_value_position == 2:
		model_based_mask = build_mask(model_based_mask, used_features)
		print(model_based_mask)

		print("Returning features selected with the model based selector!\n")
		return model_based_selector, model_based_train_features_selected, model_based_mask
	else:
		iterative_based_mask = build_mask(iterative_based_mask, used_features)
		print(iterative_based_mask)

		print("Returning features selected with the iterative based selector!\n")
		return iterative_based_selector, iterative_based_train_features_selected, iterative_based_mask



import os
import time
from gensim.models import KeyedVectors

from ..__init__ import ROOT_PATH

def load_embeddings_models():
	""" Function used to load the word-embedding models """

	# ---LOADING WORD2VEC MODEL---
	model_load_path = os.path.join(ROOT_PATH, 'models', 'word2vec', 'NILC', 'nilc_cbow_s300_300k.txt')
	start_time = time.time()
	print("Started loading the word2vec model")
	word2vec_model = KeyedVectors.load_word2vec_format(model_load_path)
	# word2vec_model = None
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')

	# ---LOADING FASTTEXT MODEL---
	model_path = os.path.join(ROOT_PATH, 'models', 'fastText', 'cc.pt.300_300k.vec')
	start_time = time.time()
	print("Started loading the fasttext model")
	fasttext_model = KeyedVectors.load_word2vec_format(model_path)
	# fasttext_model = None
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')	

	# ---LOADING PT-LKB MODEL---
	model_load_path = os.path.join(ROOT_PATH, 'models', 'ontoPT', 'PT-LKB_embeddings_64', 'ptlkb_64_30_200_p_str.emb')
	start_time = time.time()
	print("Started loading the PT-LKB-64 model")
	ptlkb64_model = KeyedVectors.load_word2vec_format(model_load_path)
	# ptlkb64_model = None
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')

	# ---LOADING GLOVE-300 MODEL---
	model_load_path = os.path.join(ROOT_PATH, 'models', 'glove', 'glove_s300_300k.txt')
	start_time = time.time()
	print("Started loading the GLOVE 300 dimensions model")
	glove300_model = KeyedVectors.load_word2vec_format(model_load_path)
	# glove300_model = None
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')

	# ---LOADING NUMBERBATCH MODEL---
	model_load_path = os.path.join(ROOT_PATH, 'models', 'numberbatch', 'numberbatch-17.02_pt_tratado.txt')
	start_time = time.time()
	print("Started loading the NUMBERBATCH dimensions model")
	numberbatch_model = KeyedVectors.load_word2vec_format(model_load_path)
	# numberbatch_model = None
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')

	return word2vec_model, fasttext_model, ptlkb64_model, glove300_model, numberbatch_model



'''
Module to compute the fasttext similarity in the ASSIN collection.
'''

import time
import os
from xml.etree import cElementTree as ET
from gensim.models import KeyedVectors
from sklearn.metrics.pairwise import cosine_similarity
#import function to read the xml file
#from assin.assineval.commons import read_xml
#import tools as tl
from ...scripts.tools import preprocessing, compute_tfidf_matrix, deprecated_read_corpus

global n_iterations
n_iterations = 0

def apply_tfidf_model(sentence, embedding_model, tfidf_model):
	""" Auxiliar function used to compute the vector of each sentence by applying the word2vec model and the tfidf model"""

	global n_iterations

	updated_sentence = []

	for word in sentence:
		if word in embedding_model.vocab:
			new_word = embedding_model[word] * tfidf_model[n_iterations][word]
			updated_sentence.append(new_word)

	n_iterations += 2
	return updated_sentence

def compute_models(model, embeddings_data, tfidf_data, use_tf_idf=0, rm_stopwords=0, numbers_to_text=0):
	""" Function used to load the FastText model and compute the vector of each sentence """

	global n_iterations

	if use_tf_idf == 1:

		# tokenization should be 0 when applying preprocessing to the date that will be used to build the tfidf matrix
		tfidf_preprocessed = preprocessing(tfidf_data, 0, rm_stopwords, numbers_to_text, use_tf_idf)
		tfidf_model = compute_tfidf_matrix(tfidf_preprocessed, rm_stopwords, 1, 0)

		n_iterations = 0
		embeddings_data['text'] = embeddings_data['text'].apply(lambda x: apply_tfidf_model(x, model, tfidf_model))
		n_iterations = 1
		embeddings_data['response'] = embeddings_data['response'].apply(lambda x: apply_tfidf_model(x, model, tfidf_model))
		n_iterations = 0
	else:
		#convert words to vectors
		embeddings_data['text'] = embeddings_data['text'].apply(lambda x: [model[word] for word in x if word in model.vocab])
		embeddings_data['response'] = embeddings_data['response'].apply(lambda x: [model[word] for word in x if word in model.vocab])

	#compute the mean of the sentence
	embeddings_data['text'] = embeddings_data['text'].apply(lambda x: sum(x)/len(x) if len(x) != 0 else [0]*300)
	embeddings_data['response'] = embeddings_data['response'].apply(lambda x: sum(x)/len(x) if len(x) != 0 else [0]*300)

	return embeddings_data

def fasttext_model(model, corpus, tf_idf, remove_stopwords, convert_num_to_text):
	fasttext_embeddings = []

	# when it comes to fasttext, the preprocessing function should always have tokenization equal to 1
	preprocessed_corpus = preprocessing(corpus, 1, remove_stopwords, convert_num_to_text, tf_idf)

	if tf_idf == 1:
		embeddings_data = compute_models(model, preprocessed_corpus, corpus, tf_idf, remove_stopwords, convert_num_to_text)
	else:
		embeddings_data = compute_models(model, preprocessed_corpus, corpus, tf_idf)

	for i in range(len(embeddings_data)):
		similarity = cosine_similarity([embeddings_data['text'][i]], [embeddings_data['response'][i]])
		fasttext_embeddings.append(similarity[0][0])

	return fasttext_embeddings

def fasttext_model_single_feature():
	""" Function used to compute similarity """

	# ---LOADING FASTTEXT MODEL---
	model_path = os.path.join('models', 'fastText', 'cc.pt.300.vec')
	start_time = time.time()
	#model = FastText.load_fasttext_format(model_path)
	#model.build_vocab(train_data, update=True)
	#model.train(train_data, epochs=model.iter, total_examples=model.corpus_count)
	print("Started loading the model")
	model = KeyedVectors.load_word2vec_format(model_path)
	print("Model loaded")
	print("--- %s seconds ---" %(time.time() - start_time))
	print('\a')

	write_path = os.path.join('assin', 'assin_results', 'fasttext_facebook_pretrained_stp_rmv_num_conv_tfidf.xml')

	test_list = read_xml("assin-ptpt-test.xml", 1)

	corpus = deprecated_read_corpus("assin-ptpt-test.xml")

	#function arguments
	remove_stopwords = 1
	convert_num_to_text = 1
	tf_idf = 1
	# when it comes to fasttext, the preprocessing function should always have tokenization equal to 1
	preprocessed_corpus = preprocessing(corpus, 1, remove_stopwords, convert_num_to_text, tf_idf)

	if tf_idf == 1:
		test_data = compute_models(model, preprocessed_corpus, corpus, tf_idf, remove_stopwords, convert_num_to_text)
	else:
		test_data = compute_models(model, preprocessed_corpus, corpus, tf_idf)

	for i in range(len(test_data)):
		similarity = cosine_similarity([test_data['text'][i]], [test_data['response'][i]])
		test_list[i].similarity = similarity[0][0]

	# write output
	tree = ET.parse("assin-ptpt-test.xml")
	root = tree.getroot()
	for i in range(len(test_list)):
		pairs = root[i]
		pairs.set('similarity', str(test_list[i].similarity))

	tree.write(write_path, 'utf-8')

#fasttext_model_single_feature()



import os

from assin.assineval.commons import read_xml
from gensim.parsing.preprocessing import strip_multiple_whitespaces

def build_training_corpus(filenames):
    corpus = []

    for filename in filenames:
        train_list = read_xml(filename, 1)

        for pair in train_list:
            corpus += [strip_multiple_whitespaces(pair.t)]
            corpus += [strip_multiple_whitespaces(pair.h)]

    return corpus

def write_training_corpus(data, filename):
	""" Function used to debug the corpus state during preprocessing """
	with open(filename, 'w') as f:
		for item in data:
			f.write("%s\n" % item)

if __name__ == "__main__":
	filenames_list = []

	corpus_path_1 = os.path.join('assin-ptpt-train.xml')
	corpus_path_2 = os.path.join('assin-ptbr-train.xml')
	corpus_path_3 = os.path.join('assin-ptpt-test.xml')
	corpus_path_4 = os.path.join('assin-ptbr-test.xml')
	corpus_path_5 = os.path.join('assin2-train.xml')

	filenames_list.extend([corpus_path_1, corpus_path_2, corpus_path_3, corpus_path_4, corpus_path_5])

	train_corpus = build_training_corpus(filenames_list)

	write_training_corpus(train_corpus, "complete_assin_training_assin1_testing_corpus.txt")
	



'''
The MIT License (MIT)

Copyright (c) 2016 Erick Rocha Fonseca

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

# -*- coding: utf-8 -*-

"""
Common structures and functions used by other scripts.
"""

from xml.etree import cElementTree as ET


str_to_entailment = {'none': 0,
                     'entailment': 1,
                     'paraphrase': 2}
entailment_to_str = {v: k for k, v in str_to_entailment.items()}

class Pair(object):
    '''
    Class representing a pair of texts from SICK or RTE.
    It is meant to be used as an abstract representation for both.
    '''
    def __init__(self, t, h, id_, entailment, similarity):
        '''
        :param t: string with the text
        :param h: string with the hypothesis
        :param id_: int indicating id in the original file
        :param entailment: int indicating entailment class
        :param similarity: float
        '''
        self.t = t
        self.h = h
        self.id = id_
        self.entailment = entailment
        self.similarity = similarity


def read_xml(filename, need_labels):
    '''
    Read an RTE XML file and return a list of Pair objects.
    :param filename: name of the file to read
    :param need_labels: boolean indicating if labels should be present
    '''
    pairs = []
    tree = ET.parse(filename)
    root = tree.getroot()

    for xml_pair in root.iter('pair'):
        t = xml_pair.find('t').text
        h = xml_pair.find('h').text
        attribs = dict(xml_pair.items())
        id_ = attribs['id']

        if 'entailment' in attribs:
            ent_string = attribs['entailment'].lower()

            try:
                ent_value = str_to_entailment[ent_string]
            except ValueError:
                msg = 'Unexpected value for attribute "entailment" at pair {}: {}'
                raise ValueError(msg.format(id_, ent_string))

        else:
            ent_value = None

        if 'similarity' in attribs:
            similarity = float(attribs['similarity'])
        else:
            similarity = None

        if need_labels and similarity is None and ent_value is None:
            msg = 'Missing both entailment and similarity values for pair {}'.format(id_)
            raise ValueError(msg)

        pair = Pair(t, h, id_, ent_value, similarity)
        pairs.append(pair)

    return pairs


def read_xml_no_attributes(filename):
    '''
    Read an RTE XML file and return a list of Pair objects.
    :param filename: name of the file to read
    :param need_labels: boolean indicating if labels should be present
    '''
    pairs = []
    tree = ET.parse(filename)
    root = tree.getroot()

    for xml_pair in root.iter('pair'):
        t = xml_pair.find('t').text
        h = xml_pair.find('h').text
        attribs = dict(xml_pair.items())
        id_ = attribs['id']

        pair = Pair(t, h, id_, None, None)
        pairs.append(pair)

    return pairs
